/*******************************************************************************
  File : evaluate.c

  Description : Module d'valuation statique d'une position.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "debug.h"
#include "evaluate.h"

/*******************************************************************************
  Constants :
*******************************************************************************/

/* Bonuses and maluses for pawns : */
const int bonus_connected_pawn[3] = { 8, 12, 16};
const int bonus_passed_pawn[3] = { 1, 2, 3 };
const int bonus_strong_square = 4;
const int malus_backward_pawn[3] = { -2, -4, -6 };
const int malus_isolated_pawn[3] = { -2, -4, -6 };
const int malus_weak_backward_pawn[3] = { -4, -4, -4 };
const int malus_weak_isolated_pawn[3] = { -8, -8, -8 };
const int malus_doubled_pawn[3] = { -1, -2, -4 };
const int malus_block_pawns[2] = { -8, -4 };

/* Bonus for bishop : */
const int bonus_bishop_fianchetto = 8;
const int bonus_bishop_pair = 4;

/* Malus for queen : */
const int malus_queen_opening = -4;

/* Bonuses for rooks : */
const int bonus_rook_half_open_file = 4;
const int bonus_rook_open_file = 4;
const int bonus_rook_seventh_rank[3] = { 4, 8, 12};
const int bonus_heavy_pieces_doubling = 2;

/* Bonuses and maluses for king : */
const int bonus_castling[2] = { 16, 24 };
const int malus_king_cover[2][3] = { { -4, -4, -4 }, { -8, -8, -8 } };
const int malus_king_no_cover[2] = { -8, -16 };
const int malus_no_castling[2] = { -4, -8 };

/* Bonus and malus to encourage, or not, exchanges in the endgame : */
const int bonus_exchange = 20;
const int malus_exchange = 10;

/* Bonus pour la position du pion (favorise l'avancement du pion) : */
const int bonus_pawn[9][9] = {
  { 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { 0, 0, 0, 1, 2, 3, 4, 5, 0 },
  { 0, 0, 0, 2, 4, 6, 8,10, 0 },
  { 0, 0, 0, 3, 6, 9,12,15, 0 },
  { 0, 0, 0, 4, 8,12,16,20, 0 },
  { 0, 0, 0, 4, 8,12,16,20, 0 },
  { 0, 0, 0, 3, 6, 9,12,15, 0 },
  { 0, 0, 0, 2, 4, 6, 8,10, 0 },
  { 0, 0, 0, 1, 2, 3, 4, 5, 0 }
};

/* Bonus pour la position du cavalier : */
const int bonus_knight[9][9] = {
  { 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { 0,-4, 0, 0, 0, 0, 0, 0,-4 },
  { 0, 0, 0, 4, 4, 4, 4, 0, 0 },
  { 0, 0, 4, 8, 8, 8, 8, 4, 0 },
  { 0, 0, 4, 8, 8, 8, 8, 4, 0 },
  { 0, 0, 4, 8, 8, 8, 8, 4, 0 },
  { 0, 0, 4, 8, 8, 8, 8, 4, 0 },
  { 0, 0, 0, 4, 4, 4, 4, 0, 0 },
  { 0,-4, 0, 0, 0, 0, 0, 0,-4 }
};

/* Bonus pour la position du fou : */
const int bonus_bishop[9][9] = {
  { 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { 0,-4, 0, 0, 0, 0, 0, 0,-4 },
  { 0, 0, 4, 4, 4, 4, 4, 4, 0 },
  { 0, 0, 4, 8, 8, 8, 8, 4, 0 },
  { 0, 0, 4, 8, 4, 4, 8, 4, 0 },
  { 0, 0, 4, 8, 4, 4, 8, 4, 0 },
  { 0, 0, 4, 8, 8, 8, 8, 4, 0 },
  { 0, 0, 4, 4, 4, 4, 4, 4, 0 },
  { 0,-4, 0, 0, 0, 0, 0, 0,-4 }
};

/* Bonus pour la position du roi (dbut de partie) : */
const int bonus_king[9][9] = {
  { 0,  0,  0,  0,  0,  0,  0,  0,  0 },
  { 0,  4,  0, -4, -8,-12,-16,-20,-24 },
  { 0, 12,  0, -4, -8,-12,-16,-20,-24 },
  { 0,  8,  0, -4, -8,-12,-16,-20,-24 },
  { 0,  0,  0, -4, -8,-12,-16,-20,-24 },
  { 0,  0,  0, -4, -8,-12,-16,-20,-24 },
  { 0,  8,  0, -4, -8,-12,-16,-20,-24 },
  { 0, 12,  0, -4, -8,-12,-16,-20,-24 },
  { 0,  4,  0, -4, -8,-12,-16,-20,-24 }
};

/* Bonus pour la centralisation du roi en finale : */
const int bonus_king_endgame[9][9] = {
  { 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { 0,-8,-4, 0, 0, 0, 0,-4,-8 },
  { 0,-4, 0, 0, 0, 0, 0, 0,-4 },
  { 0, 0, 0, 4, 4, 4, 4, 0, 0 },
  { 0, 0, 0, 4, 8, 8, 4, 0, 0 },
  { 0, 0, 0, 4, 8, 8, 4, 0, 0 },
  { 0, 0, 0, 4, 4, 4, 4, 0, 0 },
  { 0,-4, 0, 0, 0, 0, 0, 0,-4 },
  { 0,-8,-4, 0, 0, 0, 0,-4,-8 }
};

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int check_game_stage_calls = 0;
int evaluate_position_calls = 0;
#endif /* DEBUG */

#ifdef DEBUG_EVAL
int before;
#endif /* DEBUG_EVAL */

/* Pour mmoriser le nombre de fous de chaque couleur : */
int bishops_count[2];

/* Pour marquer les cases fortes : */
int bonus_square[9][9];

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : check_game_stage

  Description : Evaluation de la phase de la partie pour une couleur donne.

  Validation level : 1 - evaluate.check_game_stage
*******************************************************************************/
int check_game_stage (int color) {

/* ------------------------------- Local data ------------------------------- */

bool queens_presence = false;

int game_stage;
int x;
int pieces_move = 0;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("check_game_stage", false, &check_game_stage_calls);
#endif /* DEBUG */

  /* Si un grand nombre de pices ont t changes, */
  if (piece_count < 5) {

    /* alors la partie est dans la phase finale. */
    game_stage = ENDGAME;
  }
  else {

    /* Cas de passage en milieu de partie pour la couleur au trait :
       1) Le roque est effectu (le cas le plus frquent).
       2) Le roque n'est plus possible et au moins 3 pices (mineures + dame)
          ont t dplaces.
       3) Au moins 30 demi-coups ont t jous. */

    /* Calcule le nombre de pices qui ont boug pour la couleur au trait : */
    for (x = B; x <= G; x++) {

      if (board[x][RR(1)].moved) {

        pieces_move++;
      }
    }

    /* Si l'un des cas prcit est rencontr, */
    if (   (CASTLE_STATUS & castle_done_mask[color])
        || (   !(CASTLE_STATUS & castle_by_color_mask[color])
            && pieces_move >= 3)
        || (game_ply > 20)) {

      /* alors c'est la phase du milieu de partie, */
      game_stage = MIDDLE;
    }
    else {

      /* sinon la partie est encore dans la phase d'ouverture. */
      game_stage = OPENING;
    }
  }

  return game_stage;
}
/* End of function : check_game_stage */

/*******************************************************************************
  Function : evaluate_position

  Description : Evaluation statique d'une position.

  Validation level : 1 - evaluate.evaluate_position
*******************************************************************************/
int evaluate_position (bool debug_mode) {

/* ------------------------------- Local data ------------------------------- */

bool backward;
bool isolated;

int cf;                  /* color_factor : Pour prendre en compte la couleur. */
int color;
int gs;            /* game_stage : Phase de jeu (ouverture, milieu ou finale) */
int heavy_pieces[2][9];       /* Pour tester le doublement des pices lourdes */
int i, j;
int pawn_rank[2][10];          /* Pour mmoriser la range des pions arrirs */
int pawns[2][10];       /* Pour le comptage des pions par colonne et couleur. */
int score;
int x, y;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("evaluate_position", true, &evaluate_position_calls);
#endif /* DEBUG */

  score = 0;

  memset (bishops_count, 0, sizeof (bishops_count));
  memset (bonus_square, 0, sizeof (bonus_square));
  memset (heavy_pieces, 0, sizeof (heavy_pieces));
  memset (pawns, 0, sizeof (pawns));

  for (i = 0; i < 10; i++) {
    pawn_rank[WHITE][i] = 7;
    pawn_rank[BLACK][i] = 2;
  }

  for (color = BLACK; color <= WHITE; color++) {

    for (j = 1; j <= pieces_number[color]; j++) {

      if (!pieces[color][j].captured) {

        x = pieces[color][j].x;
        y = pieces[color][j].y;

        switch (board[x][y].piece) {

          case (PAWN) :

            pawns[color][x]++;

            if (RR(y) < RR(pawn_rank[color][x])) {

              pawn_rank[color][x] = y;
            }
            break;

          case (BISHOP) :

            bishops_count[color]++;
            break;

          case (ROOK) :

            heavy_pieces[color][x]++;
            break;

          case (QUEEN) :

            heavy_pieces[color][x]++;
            break;
        }
      }
    }
  }

#ifdef DEBUG_EVAL
  for (color = BLACK; color <= WHITE; color++) {
    for (j = 1; j <= 8; j++) {
      if (debug_mode) {
        fprintf(debug_eval_file, "pawns[%d][%d] %d - pawn_rank[%d][%d] %d\n",
          color, j, pawns[color][j], color, j, pawn_rank[color][j]);
      }
    }
  }
#endif /* DEBUG_EVAL */

  for (color = WHITE; color >= BLACK; color--) {

    cf = color ? 1 : -1;

    gs = check_game_stage (color);

    for (j = 1; j <= pieces_number[color]; j++) {

      if (!pieces[color][j].captured) {

        x = pieces[color][j].x;
        y = pieces[color][j].y;

        switch (board[x][y].piece) {

          case (PAWN) :
#ifdef DEBUG_EVAL
if (debug_mode)
before=score;
#endif
            backward = false;
            isolated = false;

            /* Bonus pour la valeur de la pice : */
            score += pieces_value[PAWN] * cf;

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> pawn value %d \n",
  color?'w':'b',x+(int)'a'-1,y,pieces_value[PAWN]*cf);
#endif
            /* Bonus pour favoriser l'avance vers la case de promotion : */
            score += bonus_pawn[x][RR(y)] * cf;

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> pawn bonus %d \n",
  color?'w':'b',x+(int)'a'-1,y,bonus_pawn[x][RR(y)] * cf);
#endif /* DEBUG_EVAL */
            /* Si les pions adjacents sont plus avancs, */
            if (   RR(pawn_rank[color][x + 1]) > RR(y)
                && RR(pawn_rank[color][x - 1]) > RR(y)) {

              /* alors le pion est arrir : */
              backward = true;

              /* et un malus est appliqu, sauf dans l'ouverture pour un pion
                 sur sa ligne de dpart : */
              if (gs != OPENING || RR(y) != 2) {

                score += malus_backward_pawn[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> malus_backward_pawn %d \n",
  color?'w':'b',x+(int)'a'-1,y,malus_backward_pawn[gs] * cf);
#endif /* DEBUG_EVAL */
              }

              /* La case situe devant le pion est une case forte pour
                 l'adversaire : */
              bonus_square[x][y + 1 * cf] -= bonus_strong_square * cf;

              /* Si les pions adjacents ont disparus, */
              if (   !pawns[color][x + 1]
                  && !pawns[color][x - 1]) {

                /*  alors le pion est isol : */
                isolated = true;

                /* et un malus supplmentaire est appliqu. */
                score += malus_isolated_pawn[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> malus_isolated_pawn %d \n",
  color?'w':'b',x+(int)'a'-1,y,malus_isolated_pawn[gs] * cf);
#endif /* DEBUG_EVAL */
              }
            }

            /* Si le pion est sur une colonne semi-ouverte, */
            if (!pawns[color^1][x]) {

              /* alors un malus est appliqu si le pion est arrir : */
              if (backward) {

                score += malus_weak_backward_pawn[gs] * cf;
#ifdef DEBUG_EVAL
fprintf(debug_eval_file, "gs %d pawn[%d][%d] %d (malus_weak_backward_pawn)\n",
  gs, x, y, score - before);
#endif
              }

              /* avec un malus supplmentaire s'il est isol : */
              if (isolated) {

                score += malus_weak_isolated_pawn[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> malus_weak_isolated_pawn %d \n",
  color?'w':'b',x+(int)'a'-1,y,malus_weak_isolated_pawn[gs] * cf);
#endif /* DEBUG_EVAL */
              }
            }

            /* Les pions doubls, voire trippls, sont lourdement pnaliss : */
            if (pawns[color][x] > 1) {

              score += malus_doubled_pawn[gs] * cf * (pawns[color][x] - 1);
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> malus_doubled_pawn %d \n",
  color?'w':'b',x+(int)'a'-1,y,malus_doubled_pawn[gs] * cf*(pawns[color][x]-1));
#endif /* DEBUG_EVAL */
            }

            /* Les pions passs (aucun pion adverse en face et sur les colonnes
               adjacentes) sont favoriss, et ce d'autant plus qu'ils sont
               avancs : */
            if (   !pawns[color^1][x]
                && RR(y) >= RR(pawn_rank[color^1][x - 1])
                && RR(y) >= RR(pawn_rank[color^1][x + 1])) {

              score += bonus_passed_pawn[gs] * cf * bonus_pawn[x][RR(y)];
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> bonus_passed_pawn %d \n",
  color?'w':'b',x+(int)'a'-1,y,bonus_passed_pawn[gs] * cf*bonus_pawn[x][RR(y)]);
#endif /* DEBUG_EVAL */

              /* Avec un bonus supplmentaire si ils sont soutenus : */
              if (   !isolated
                  && (   (RR(pawn_rank[color][x]) -
                          RR(pawn_rank [color][x + 1]) == 1)
                      || (RR(pawn_rank[color][x]) -
                          RR(pawn_rank [color][x - 1]) == 1))) {

                score += bonus_connected_pawn[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> bonus_connected_pawn %d \n",
  color?'w':'b',x+(int)'a'-1,y,bonus_connected_pawn[gs] * cf);
#endif /* DEBUG_EVAL */
              }
            }
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c pawn   %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
            break;

          case (KNIGHT) :
#ifdef DEBUG_EVAL
if (debug_mode)
before=score;
#endif /* DEBUG_EVAL */
            /* Bonus pour la valeur de la pice : */
            score += pieces_value[KNIGHT] * cf;

            /* Bonus pour la position du cavalier : */
            score += bonus_knight[x][y] * cf;

            /* Bonus si le cavalier occupe une case forte : */
            score += bonus_square[x][y];

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c knight %c%d -> score %d\n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
            break;

          case (BISHOP) :
#ifdef DEBUG_EVAL
if (debug_mode)
before=score;
#endif /* DEBUG_EVAL */
            /* Bonus pour la valeur de la pice : */
            score += pieces_value[BISHOP] * cf;

            /* Bonus pour la position du fou : */
            score += bonus_bishop[x][y] * cf;

            /* Bonus si le fou occupe une case forte : */
            score += bonus_square[x][y];

            if (gs != ENDGAME) {

              /* Bonus pour le fou en fianchetto : */
              score += bonus_bishop_fianchetto * cf;
           }

           if (   gs != OPENING
               && bishops_count[color] == 2) {

              /* Bonus pour la paire de fous en milieu et fin de partie : */
              score += bonus_bishop_pair * cf;
           }

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c bishop %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
            break;

          case (ROOK) :
#ifdef DEBUG_EVAL
if (debug_mode)
before=score;
#endif /* DEBUG_EVAL */
            /* Bonus pour la valeur de la pice : */
            score += pieces_value[ROOK] * cf;

            /* La septime range est intressante pour une tour : */
            if (RR(y) == 7) {

              score += bonus_rook_seventh_rank[gs] * cf;
            }

            /* Accorde un bonus pour une tour, */
            if (!pawns[color][x]) {

              /*  sur une colonne semi-ouverte :  */
              score += bonus_rook_half_open_file * cf;

              if (!pawns[color^1][x]) {

                /* et mieux sur une colonne ouverte :  */
                score += bonus_rook_open_file * cf;
              }
            }

            /* Bonus pour le doublement des pices lourdes sur une mme
               colonne : */
            if (heavy_pieces[color][x] > 1) {

              score += bonus_heavy_pieces_doubling * cf;
            }

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c rook   %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
            break;

          case (QUEEN) :
#ifdef DEBUG_EVAL
if (debug_mode)
before=score;
#endif /* DEBUG_EVAL */
            /* Bonus pour la valeur de la pice : */
            score += pieces_value[QUEEN] * cf;

            if (gs == OPENING) {

              /* Le dplacement de la dame dans l'ouverture avant les pices
                 mineures est pnalis : */
              if (   board[D][RR(1)].moved) {

                if (   !board[B][RR(1)].moved || !board[C][RR(1)].moved
                    || !board[F][RR(1)].moved || !board[G][RR(1)].moved) {

                  score += malus_queen_opening * cf;
                }
              }
            }

            /* Bonus pour le doublement des pices lourdes sur une mme
               colonne : */
            if (heavy_pieces[color][x] > 1) {

              score += bonus_heavy_pieces_doubling * cf;
            }

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c queen  %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
            break;

          case (KING) :
#ifdef DEBUG_EVAL
if (debug_mode) {
  before = score;
fprintf(debug_eval_file, "gs %d king[%d][%d] %d (%x %d %d)\n", gs, x, y, score,
  CASTLE_STATUS,
  CASTLE_STATUS & castle_done_mask[color],
  CASTLE_STATUS & castle_by_color_mask[color]);
}
#endif
            if (gs == ENDGAME) {

              /* Bonus pour la centralisation en finale : */
              score += bonus_king_endgame[x][y] * cf;

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kb end %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
            }
            else {

              /* Bonus pour la position du roi : */
              score += bonus_king[x][RR(y)] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kb     %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */

              /* Si le roi a roqu, */
              if (CASTLE_STATUS & castle_done_mask[color]) {

                /* alors il obtient un bonus : */
                score += bonus_castling[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kbOO   %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
              }
              /* par contre s'il ne peut plus roquer, */
              else if (!(CASTLE_STATUS & castle_by_color_mask[color])) {

                /* alors il obtient un malus : */
                score += malus_no_castling[gs] * cf;

#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kmnoOO %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
             }

              /* Pnalit si la protection du roi, par les pions situs 
                devant lui, est dtriore : */
              if (pawns[color][x] && RR(y) < RR(pawn_rank[color][x])) {

                score +=   malus_king_cover[gs][1] * cf
                         * (RR(pawn_rank[color][x]) - RR(y) - 1);
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kmcff  %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
              }
              else {

                score += malus_king_no_cover[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kmcnf  %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
              }

              if (pawns[color][x + 1] && RR(y) < RR(pawn_rank[color][x + 1])) {

                score +=   malus_king_cover[gs][2] * cf
                         * (RR(pawn_rank[color][x + 1]) - RR(y) - 1);
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kmcf+  %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
              }
              else {

                score += malus_king_no_cover[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kmcn+  %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
              }

              if (pawns[color][x - 1] && RR(y) < RR(pawn_rank[color][x - 1])) {

                score +=   malus_king_cover[gs][0] * cf
                         * (RR(pawn_rank[color][x - 1]) - RR(y) - 1);
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kmcf-  %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
              }
              else {

                score += malus_king_no_cover[gs] * cf;
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c kmcn-  %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
              }
            }
#ifdef DEBUG_EVAL
if (debug_mode)
fprintf(debug_eval_file, "DBG - %c king   %c%d -> score %d \n",
  color?'w':'b',x+(int)'a'-1,y,score-before);
#endif /* DEBUG_EVAL */
            break;
        }
      }
    }

    if (gs == OPENING || gs == MIDDLE) {

#ifdef DEBUG_EVAL
before = score;
fprintf(debug_eval_file, "gs %d block %d\n", gs, score);
#endif
      /* Pnalise le blocage des pions centraux dans le dbut de partie : */
      if (!board[D][RR(2)].moved && board[D][RR(3)].piece) {

        score += malus_block_pawns[gs] * cf;
      }

      if (!board[E][RR(2)].moved && board[E][RR(3)].piece) {

        score += malus_block_pawns[gs] * cf;
      }
    }
#ifdef DEBUG_EVAL
fprintf(debug_eval_file, "gs %d block %d\n", gs, score - before);
#endif
  }

#ifdef DEBUG_EVAL
before = score;
fprintf(debug_eval_file, "gs %d trade %d\n", gs, score);
#endif

  if (gs == ENDGAME) {

    /* En finale, favorise l'change de matriel si le programme possde un bon
       avantage matriel : */
    if (score * cf > 150 && piece_count < start_piece_count) {

      score += bonus_exchange * cf;
    }
    /* dans le cas contraire les changes sont vits : */
    else if (score * cf < 80 && piece_count < start_piece_count) {

      score += malus_exchange * cf;
    }
  }

#ifdef DEBUG_EVAL
fprintf(debug_eval_file, "gs %d trade %d\n", gs, score - before);
#endif

#ifdef DEBUG_EVAL
fprintf(debug_eval_file, "active_color %d - score %d\n",
  active_color, active_color == BLACK ? - score : score);
#endif

#ifdef DEBUG_HIGH
debug.print (" : active_color %d - score %d\n",
  active_color, active_color == BLACK ? - score : score);
#endif /* DEBUG_HIGH */

  return (active_color == BLACK ? - score : score);
}
/* End of function : evaluate_position */

/*******************************************************************************
  Object declaration :
*******************************************************************************/
evaluate_module_s evaluate = {

  check_game_stage,
  evaluate_position
};
