/*******************************************************************************
  File : hash.c

  Description : Module de hachage.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "debug.h"
#include "hash.h"
#include "moves.h"
#include "tools.h"
#include "uci.h"

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int compute_h_position_calls = 0;
int init_h_values_calls = 0;
#endif /* DEBUG */

/* Valeur de hachage des positions de la partie :
   (Pour vrifier la rptition des position) */
h_value_s h_position;
h_value_s h_position_history[MOVES_LIMIT * 2];

/* Table des valeurs de hachage pour successivement les pices, la couleur au
   trait, la prise en passant et les roques : */
h_value_s h_values[2][17][9][9];
h_value_s h_values_color[2];
h_value_s h_values_ep[9][9];
h_value_s h_values_castle[5][2];

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : compute_h_position

  Description : Calcule une valeur de hachage pour la position courante. C'est 
                une valeur numrique unique pour une position donne.

  Validation level : 1 - hash.compute_h_position
*******************************************************************************/
h_value_s compute_h_position (void) {

/* ------------------------------- Local data ------------------------------- */

h_value_s h_current_position = 0;

int color;
int i;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("compute_h_position\n", true, &compute_h_position_calls);
#endif /* DEBUG */

  /* Prise en compte des pices prsentes sur l'chiquier, de leur couleur et
     de leur position : */
  for (color = BLACK; color <= WHITE; color++) {

    for (i = 1; i <= pieces_number[color]; i++) {

      if (pieces[color][i].piece) {

        h_current_position ^= h_values[color][pieces[color][i].piece]
                                             [pieces[color][i].x]
                                             [pieces[color][i].y];
      }
    }
  }

  /* Prise en compte de la case de prise en passant ventuelle : */
  h_current_position ^= h_values_ep[EP_SQUARE.x][EP_SQUARE.y];

  /* Prise en compte des roques :
    (voir la description de castle_status dans data.c) */
  for (i = 1; i <= 4; i++) {

    h_current_position ^=
      h_values_castle[i][CASTLE_STATUS & castle_data_mask[i]];
  }

  /* Prise en compte de la couleur au trait : */
  h_current_position ^= h_values_color[active_color];

  return (h_current_position);
}
/* End of function : compute_h_position */

/*******************************************************************************
  Function : init_h_values

  Description : Calcul des valeurs de hachage pour tous les lments qui
                permettent de caractriser une position.

  Validation level : 1 - hash.init_h_values
*******************************************************************************/
void init_h_values (void) {

/* ------------------------------- Local data ------------------------------- */

int color;
int i;
int piece;
int x, y;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("init_h_values\n", true, &init_h_values_calls);
#endif /* DEBUG */

  /* Efface le contenu des tables des valeurs de hashage : */
  memset (h_values, 0, sizeof (h_values));
  memset (h_values_ep, 0, sizeof (h_values_ep));

  for (color = BLACK; color <= WHITE; color++) {

    for (piece = PAWN; piece <= KING; piece++) {

      for (x = A; x <= H; x++) {

        for (y = 1; y <= 8; y++) {

          /* Calcul des valeurs de hashage pour les pices : */
          h_values[color][piece][x][y] = tools.generate_random ();
        }

        /* Calcul des valeurs de hashage pour la prise en passant : */
        h_values_ep[x][RR(3)] = tools.generate_random ();
      }
    }

    /* Calcul des valeurs de hashage pour la couleur au trait : */
    h_values_color[color] = tools.generate_random ();
  }

  /* Calcul des valeurs de hachage pour la prise en compte des roques :
     (fonction de sa possibilit) : */
  for (i = 1; i <= 4; i++) {

    h_values_castle[i][0] = tools.generate_random ();
    h_values_castle[i][1] = tools.generate_random ();
  }
}
/* End of function : init_h_values */

/*******************************************************************************
  Module declaration :
*******************************************************************************/
hash_module hash = {

  compute_h_position,
  init_h_values
};
