/*******************************************************************************
  File : moves.c

  Description : Module de gnration des coups et de dplacement des pices.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "debug.h"
#include "hash.h"
#include "moves.h"
#include "tools.h"

/*******************************************************************************
  Types :
*******************************************************************************/

/* Informations qui dcrivent un dplacement : */
typedef struct {

  /* Dplacement en nombres de colonnes : */
  int vx;

  /* Dplacement en nombres de rangs : */
  int vy;

} vector_s;

/* Informations sur les dplacements possibles d'un pion : */
typedef struct {

  /* Liste des directions que peut prendre un pion : */
  int direction[5];

} pawn_liberty_s;

/* Informations sur les dplacements possibles d'une pice : */
typedef struct {

  /* Indique une pice  long rayon d'action : */
  bool long_range;

  /* Liste des directions que peut prendre une pice : */
  int direction[9];

} piece_liberty_s;

/*******************************************************************************
  Constants :
*******************************************************************************/

/* Vecteurs de dplacement dans chacune des 18 directions : */
const vector_s vector[19] = {
  { 0, 0}, /* Non utilise. */

  /* Dplacement d'une case dans toutes les directions : */
  { 0, 1}, { 1, 1}, { 1, 0}, { 1,-1}, { 0,-1}, {-1,-1}, {-1, 0}, {-1, 1},

  /* Pour les pions  partir de leur case de dpart : */
  { 0, 2}, { 0, -2},

  /* Pour les cavaliers : */
  { 1, 2}, { 2, 1}, { 2,-1}, { 1,-2}, {-1,-2}, {-2,-1}, {-2, 1}, {-1, 2}};

/* Directions autorises pour les pions : */
const pawn_liberty_s pawn_liberty[2] = {
  { 0, 5,10, 4, 6 }, /* BLACK PAWN */
  { 0, 1, 9, 2, 8 }  /* WHITE PAWN */
};

/* Directions autorises pour les pices (sauf les pions) : */
const piece_liberty_s liberty[7] = {
  { false,  0,  0,  0,  0,  0,  0,  0,  0,  0 }, /* NOT USED */
  { false,  0,  0,  0,  0,  0,  0,  0,  0,  0 }, /* NOT USED */

  { false,  0, 11, 12, 13, 14, 15, 16, 17, 18 }, /* KNIGHT */
  {  true,  0,  0,  2,  0,  4,  0,  6,  0,  8 }, /* BISHOP */
  {  true,  0,  1,  0,  3,  0,  5,  0,  7,  0 }, /* ROOK */
  {  true,  0,  1,  2,  3,  4,  5,  6,  7,  8 }, /* QUEEN */
  { false,  0,  1,  2,  3,  4,  5,  6,  7,  8 }  /* KING */
};

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int generate_calls = 0;
int search_auth_dir_calls = 0;
int is_attacked_calls = 0;
int make_calls = 0;
int scan_board_calls = 0;
int unmake_calls = 0;
#endif /* DEBUG */

/*******************************************************************************
  Internal functions prototypes :
*******************************************************************************/

void search_auth_dir (void);

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : generate

  Description : Gnrateur des coups lgaux  partir de la position courante et
                pour la couleur donne en entre.

  Validation level : 1 - moves.generate
*******************************************************************************/
int generate (move_s *legal_moves, int search_type) {

/* ------------------------------- Local data ------------------------------- */

bool look_forward;

int color = active_color;
int i;
int j;
int moves_number;
int tmp_moves_number;
int x;
int y, y1;

int DIRECTION;
int PAWN_DIRECTION;
int PIECES_AUTHORIZED;
int PIECES_PIECE;
int PIECES_X;
int PIECES_Y;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("generate", true, &generate_calls);
#endif /* DEBUG */

#ifdef DEBUG_HIGH
  debug.print (" : type of search %d\n", search_type);
  if (nodes >= DEBUG_NODES_MIN && nodes < DEBUG_NODES_MAX) {
    tools.print_board (debug_file);
  }
#endif /* DEBUG_HIGH */

  moves_number = 0;

  /* Marquage des pices  mobilit rduite (clouages) : */
  search_auth_dir ();

  for (i = 1; i <= pieces_number[color]; i++) {

    PIECES_PIECE = pieces[color][i].piece;
    PIECES_X = pieces[color][i].x;
    PIECES_Y = pieces[color][i].y;
    PIECES_AUTHORIZED = pieces[color][i].authorized;

    if (PIECES_PIECE != PAWN && !pieces[color][i].captured) {

      for (j = 1; j <= 8; j++) {

        DIRECTION = liberty[PIECES_PIECE].direction[j];

        if (   DIRECTION
            && (   PIECES_AUTHORIZED == -1
                || PIECES_AUTHORIZED == DIRECTION
                || OPPOSED(PIECES_AUTHORIZED) == DIRECTION)) {

          x = PIECES_X;
          y = PIECES_Y;

          look_forward = true;
          do {

            x += vector[DIRECTION].vx;
            y += vector[DIRECTION].vy;

            if (IS_INSIDE(x, y)) {

              if (   (board[x][y].piece && board[x][y].color == color)
                  || (board[x][y].piece == KING)
                  || (PIECES_PIECE == KING &&
                        moves.is_attacked (x, y, color^1) )) {

                look_forward = false;
              }
              else {

                if (search_type != CAPTURES_ONLY || board[x][y].piece) {

                  moves_number++;
                  legal_moves[moves_number].piece = PIECES_PIECE;
                  legal_moves[moves_number].color =
                    board[PIECES_X][PIECES_Y].color;
                  legal_moves[moves_number].from.x = PIECES_X;
                  legal_moves[moves_number].from.y = PIECES_Y;
                  legal_moves[moves_number].to.x = x;
                  legal_moves[moves_number].to.y = y;
                  legal_moves[moves_number].form = i;
                  legal_moves[moves_number].status = 0;
                  legal_moves[moves_number].promotion = 0;

                  if (board[x][y].piece) {

                    look_forward = false;

                    legal_moves[moves_number].captured_form = board[x][y].form;
                  }
                  else {

                    legal_moves[moves_number].captured_form = 0;
                  }
                }
              }
            }
            else {

              look_forward = false;
            }

          }  while (look_forward && liberty[PIECES_PIECE].long_range);
        }
      }
    }
    else if (PIECES_PIECE == PAWN && !pieces[color][i].captured) {

      y1 = PIECES_Y + vector[pawn_liberty[color].direction[1]].vy;

      for (j = 1; j <= 4; j++) {

        PAWN_DIRECTION = pawn_liberty[color].direction[j];

        if (   PIECES_AUTHORIZED == -1
            || PIECES_AUTHORIZED == PAWN_DIRECTION
            || (PIECES_AUTHORIZED == 1 && PAWN_DIRECTION == 9)
            || (PIECES_AUTHORIZED == 5 && PAWN_DIRECTION == 10)) {

          x = PIECES_X + vector[PAWN_DIRECTION].vx;
          y = PIECES_Y + vector[PAWN_DIRECTION].vy;

          if (IS_INSIDE(x, y)) {

            if (   (j == 1 && !board[x][y].piece)
                || (j > 2 && board[x][y].piece && board[x][y].color != color)
                || (j > 2 && EP_SQUARE.x == x && EP_SQUARE.y == y)
                || (j == 2 && !board[x][y].piece && !board[x][y1].piece
                    && PIECES_Y == RR(2))) {

              if (   (search_type != CAPTURES_ONLY)
                  || (board[x][y].piece)
                  || (j > 2 && EP_SQUARE.x == x && EP_SQUARE.y == y)
                  || (y == RR(8))) {

                moves_number++;

                legal_moves[moves_number].piece = PIECES_PIECE;
                legal_moves[moves_number].color =
                  board[PIECES_X][PIECES_Y].color;
                legal_moves[moves_number].from.x = PIECES_X;
                legal_moves[moves_number].from.y = PIECES_Y;
                legal_moves[moves_number].to.x = x;
                legal_moves[moves_number].to.y = y;
                legal_moves[moves_number].form = i;

                if (board[x][y].piece) {

                  legal_moves[moves_number].captured_form = board[x][y].form;
                }
                else {

                  legal_moves[moves_number].captured_form = 0;
                }

                /* Promotion : */
                if (y == RR(8)) {

                  legal_moves[moves_number].status = PROMOTION;
                  legal_moves[moves_number].promotion = QUEEN;

                  moves_number++;
                  memcpy (&legal_moves[moves_number],
                          &legal_moves[moves_number - 1], sizeof(move_s));
                  legal_moves[moves_number].promotion = ROOK;

                  moves_number++;
                  memcpy (&legal_moves[moves_number],
                          &legal_moves[moves_number - 1], sizeof(move_s));
                  legal_moves[moves_number].promotion = KNIGHT;

                  moves_number++;
                  memcpy (&legal_moves[moves_number],
                          &legal_moves[moves_number - 1], sizeof(move_s));
                  legal_moves[moves_number].promotion = BISHOP;
                }
                else if (j > 2 && EP_SQUARE.x == x && EP_SQUARE.y == y) {

                  legal_moves[moves_number].status = EN_PASSANT;
                  legal_moves[moves_number].captured_form =
                    board[x][(y == 3 ? 4 : 5)].form;
                }
                else {

                  legal_moves[moves_number].status = 0;
                  legal_moves[moves_number].promotion = 0;
                }
              }
            }
          }
        }
      }
    }
  }

  if (search_type != CAPTURES_ONLY) {

    /* Recherche les roques possibles : */
    if (color == WHITE) {

      if (     CASTLE_STATUS & 0x1
          && !board[F][1].piece
          && !board[G][1].piece
          && !moves.is_attacked (E, 1, BLACK)
          && !moves.is_attacked (F, 1, BLACK)
          && !moves.is_attacked (G, 1, BLACK)) {

        moves_number++;

        legal_moves[moves_number].piece = KING;
        legal_moves[moves_number].color = WHITE;
        legal_moves[moves_number].from.x = E;
        legal_moves[moves_number].from.y = 1;
        legal_moves[moves_number].to.x = G;
        legal_moves[moves_number].to.y = 1;
        legal_moves[moves_number].status = KING_CASTLE;
        legal_moves[moves_number].promotion = 0;
        legal_moves[moves_number].form = board[E][1].form;
        legal_moves[moves_number].captured_form = 0;
      }

      if (     CASTLE_STATUS & 0x2
          && !board[D][1].piece
          && !board[C][1].piece
          && !board[B][1].piece
          && !moves.is_attacked (E, 1, BLACK)
          && !moves.is_attacked (D, 1, BLACK)
          && !moves.is_attacked (C, 1, BLACK)) {

        moves_number++;

        legal_moves[moves_number].piece = KING;
        legal_moves[moves_number].color = WHITE;
        legal_moves[moves_number].from.x = E;
        legal_moves[moves_number].from.y = 1;
        legal_moves[moves_number].to.x = C;
        legal_moves[moves_number].to.y = 1;
        legal_moves[moves_number].status = QUEEN_CASTLE;
        legal_moves[moves_number].promotion = 0;
        legal_moves[moves_number].form = board[E][1].form;
        legal_moves[moves_number].captured_form = 0;
      }
    }
    else {

      if (     CASTLE_STATUS & 0x4
          && !board[F][8].piece
          && !board[G][8].piece
          && !moves.is_attacked (E, 8, WHITE)
          && !moves.is_attacked (F, 8, WHITE)
          && !moves.is_attacked (G, 8, WHITE)) {

        moves_number++;

        legal_moves[moves_number].piece = KING;
        legal_moves[moves_number].color = BLACK;
        legal_moves[moves_number].from.x = E;
        legal_moves[moves_number].from.y = 8;
        legal_moves[moves_number].to.x = G;
        legal_moves[moves_number].to.y = 8;
        legal_moves[moves_number].status = KING_CASTLE;
        legal_moves[moves_number].promotion = 0;
        legal_moves[moves_number].form = board[E][8].form;
        legal_moves[moves_number].captured_form = 0;
      }

      if (     CASTLE_STATUS & 0x8
          && !board[D][8].piece
          && !board[C][8].piece
          && !board[B][8].piece
          && !moves.is_attacked (E, 8, WHITE)
          && !moves.is_attacked (D, 8, WHITE)
          && !moves.is_attacked (C, 8, WHITE)) {

        moves_number++;

        legal_moves[moves_number].piece = KING;
        legal_moves[moves_number].color = BLACK;
        legal_moves[moves_number].from.x = E;
        legal_moves[moves_number].from.y = 8;
        legal_moves[moves_number].to.x = C;
        legal_moves[moves_number].to.y = 8;
        legal_moves[moves_number].status = QUEEN_CASTLE;
        legal_moves[moves_number].promotion = 0;
        legal_moves[moves_number].form = board[E][8].form;
        legal_moves[moves_number].captured_form = 0;
      }
    }
  }

  if (search_type == LEGAL_MOVES || search_type == CAPTURES_ONLY) {

    /* A partir des coups obtenus, on ne garde que ceux qui ne laissent pas le
     roi en chec : */
    if (moves.is_attacked (king[color].x, king[color].y, color^1)) {

      tmp_moves_number = 0;

      for (i = 1; i <= moves_number; i++) {

        moves.make (legal_moves[i]);
        if (!moves.is_attacked (king[color].x, king[color].y, color^1)) {

          tmp_moves_number++;
          legal_moves[tmp_moves_number] = legal_moves[i];
        }

        moves.unmake (legal_moves[i]);
      }

      moves_number = tmp_moves_number;
    }
  }

#ifdef DEBUG_HIGH
debug.print ("moves (%d) :", moves_number);
for (i = 1; i <= moves_number; i++) {
  debug.print (" %c%d%c%d%c",
    legal_moves[i].from.x + (int) 'a' -1, legal_moves[i].from.y,
    legal_moves[i].to.x + (int) 'a' -1, legal_moves[i].to.y,
    legal_moves[i].status == PROMOTION ? PIECE(legal_moves[i].promotion) : ' ');
}
debug.print ("\n");
#endif /* DEBUG_HIGH */

  return moves_number;
}
/* End of function : generate */

/*******************************************************************************
  Function : is_attacked

  Description : Returns true if a square (x0, y0) is attacked by a piece of the
                color given, false otherwise.

  Validation level : 1 - moves.is_attacked
*******************************************************************************/
bool is_attacked (int x0, int y0, int color) {

/* ------------------------------- Local data ------------------------------- */

bool attacked = false;
bool look_forward;

int d;
int x, y;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("is_attacked", false, &is_attacked_calls);
#endif /* DEBUG */

#ifdef DEBUG_HIGH1
  debug.print (" : %c%d by %s",
               x0 + (int) 'a' - 1, y0, (color ? "white" : "black"));
#endif /* DEBUG_HIGH */

  for (d = 1; d < 19; (d == 8) ? d += 3 : d++) {

    x = x0;
    y = y0;

    look_forward = true;
    while (look_forward) {

      x += vector[d].vx;
      y += vector[d].vy;

      if (   !IS_INSIDE(x, y)
          || (!board[x][y].piece && d > 10)
          || (board[x][y].piece && board[x][y].color != color)) {

        look_forward = false;
      }
      else {

        if (board[x][y].piece && board[x][y].color == color) {

         look_forward = false;

          if (board[x][y].piece == KNIGHT && d > 10) {

            attacked = true;
          }
          else if (d < 9) {

            if (liberty[board[x][y].piece].direction[d] == d) {

              if (   (   DISTANCE(x0, y0, x, y) == 1
                      && liberty[board[x][y].piece].long_range == false)
                  || (liberty[board[x][y].piece].long_range == true)) {

                attacked = true;
              }
            }

            if (   (board[x][y].piece == PAWN)
                && (   (pawn_liberty[color].direction[3] == OPPOSED(d))
                    || (pawn_liberty[color].direction[4] == OPPOSED(d)))
                && (DISTANCE(x0, y0, x, y) == 1)) {

              attacked = true;
            }
          }
        }
      }
    }
  }

#ifdef DEBUG_HIGH1
  debug.print (" = %d\n", attacked);
#endif /* DEBUG_HIGH */

  return attacked;
}
/* End of function : is_attacked */

/*******************************************************************************
  Function : make

  Description : Joue le coup donn en entre et met  jour les diffrentes
                donnes du programme.

  Validation level : 1 - moves.make
*******************************************************************************/
void make (move_s move) {

/* ------------------------------- Local data ------------------------------- */

int i;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("make", false, &make_calls);
#endif /* DEBUG */

#ifdef DEBUG_HIGH
  debug.print ("(%9d) - %20s", nodes, "make");
  debug.print (" : %d) %s%c%d%c%d%c\n",
    game_ply, active_color ? "" : " ..... ",
    move.from.x + (int) 'a' - 1, move.from.y,
    move.to.x + (int) 'a' - 1, move.to.y,
    move.status == PROMOTION ? PIECE(move.promotion) : ' ');
  debug.print (" make history :");
  for (i = 0; i < game_ply; i++) {
    debug.print (" %c%d%c%d%c",
      played_moves[i].from.x + (int) 'a' - 1, played_moves[i].from.y,
      played_moves[i].to.x + (int) 'a' - 1, played_moves[i].to.y,
      played_moves[i].status == PROMOTION ? PIECE(played_moves[i].promotion) : ' ');
  }
  debug.print ("\n");
#endif /* DEBUG_HIGH */

  /* Mmorise le coup dans la liste des coups jous : */
  memcpy (&played_moves[game_ply], &move, sizeof (move_s));

  h_position_history[game_ply++] = h_position;

  CASTLE_STATUS = castle_status[game_ply - 1];
  EP_SQUARE = ep_square[game_ply - 1];

  /* Mise  jour des informations de hachage pour la position courante : */
  h_position ^= h_values[move.color][move.piece][move.from.x][move.from.y];
  h_position ^= h_values[board[move.to.x][move.to.y].color]
                        [board[move.to.x][move.to.y].piece]
                        [move.to.x][move.to.y];
  h_position ^= h_values[move.color][move.piece][move.to.x][move.to.y];
  h_position ^= h_values_ep[EP_SQUARE.x][EP_SQUARE.y];
  h_position ^= h_values_color[BLACK];
  h_position ^= h_values_color[WHITE];

  /* Prise en compte des possibilits de roquer avant le coup jou : */
  for (i = 1; i <= 4; i++) {

    h_position ^= h_values_castle[i][CASTLE_STATUS & castle_data_mask[i]];
  }

  /* Efface la possibilit de prendre en passant : */
  EP_SQUARE.x = 0;
  EP_SQUARE.y = 0;

  /* Mise  jour du comptage de demi-coups pour la rgle des 50 coups : */
  halfmove_history[search_ply] = halfmove;
  if (move.piece != PAWN && !move.captured_form) {

    halfmove++;
  }
  else {

    halfmove = 0;
  }

  /* Mise  jour du dcompte des pices : */
  if (board[move.to.x][move.to.y].piece > PAWN &&
      board[move.to.x][move.to.y].piece < KING) {

    piece_count--;
  }

  /* Mise  jour des informations pour la case d'arrive : */
  board[move.to.x][move.to.y].piece = move.piece;
  board[move.to.x][move.to.y].color = move.color;
  board[move.to.x][move.to.y].form  = move.form;
  board[move.to.x][move.to.y].moved++;

  /* Mise  jour de la fiche associe : */
  pieces[move.color][move.form].x = move.to.x;
  pieces[move.color][move.form].y = move.to.y;

  /* Si c'est une capture, */
  if (move.captured_form) {

    /* alors mise  jour de la fiche associe : */
    pieces[move.color^1][move.captured_form].captured = true;
  }

  /* Effacement des informations pour la case de dpart : */
  board[move.from.x][move.from.y].piece = NONE;
  board[move.from.x][move.from.y].moved++;

  if (move.piece == PAWN) {

    /* Si c'est une promotion, la nature de la pice est change : */
    if (move.status == PROMOTION) {

      /* La pice promue est mise sur l'chiquier : */
      board[move.to.x][move.to.y].piece = move.promotion;
      pieces[move.color][move.form].piece = move.promotion;

      h_position ^= h_values[move.color][move.piece][move.to.x][move.to.y];
      h_position ^= h_values[board[move.to.x][move.to.y].color]
                            [move.promotion][move.to.x][move.to.y];
    }
    else if (move.status == EN_PASSANT) {

      /* Effacement des informations du pion captur : */
      board[move.to.x][move.from.y].piece = NONE;

      h_position ^= h_values[move.color^1][PAWN][move.to.x][move.from.y];
    }

    else if (abs (move.to.y - move.from.y) == 2) {

      /* Positionne la case pour une ventuelle prise en passant : */
      EP_SQUARE.x = move.from.x;
      EP_SQUARE.y = reverse_rank[move.color][3];

      h_position ^= h_values_ep[EP_SQUARE.x][EP_SQUARE.y];
    }
  }
  /* Si la pice dplace est un roi, */
  else if (move.piece == KING) {

    /* alors on mmorise sa nouvelle position : */
    king[move.color].x = move.to.x;
    king[move.color].y = move.to.y;

    /* Supprime les possibilits de roquer pour le roi : */
    CASTLE_STATUS &= castle_by_color_mask[move.color^1];

    /* Si c'est un petit roque la tour 'H' est dplace galement : */
    if (move.status == KING_CASTLE) {

      /* Indique que le roi a roqu : */
      CASTLE_STATUS |= castle_done_mask[move.color];

      /* Mise  jour des informations de la case d'arrive : */
      board[F][move.to.y].piece = ROOK;
      board[F][move.to.y].color = move.color;
      board[F][move.to.y].form  = board[H][move.to.y].form;

      /* Mise  jour de la fiche associe : */
      pieces[move.color][board[F][move.to.y].form].x = F;

      /* Effacement des informations de la case de dpart : */
      board[H][move.to.y].piece = NONE;

      h_position ^= h_values[move.color][ROOK][H][move.from.y];
      h_position ^= h_values[move.color][ROOK][F][move.from.y];
    }
    /* Si c'est un grand roque la tour 'A' est dplace galement : */
    else if (move.status == QUEEN_CASTLE) {

      /* Indique que le roi a roqu : */
      CASTLE_STATUS |= castle_done_mask[move.color];

      /* Mise  jour des informations de la case d'arrive : */
      board[D][move.to.y].piece = ROOK;
      board[D][move.to.y].color = move.color;
      board[D][move.to.y].form  = board[A][move.to.y].form;

      /* Mise  jour de la fiche associe : */
      pieces[move.color][board[D][move.to.y].form].x = D;

      /* Effacement des informations de la case de dpart : */
      board[A][move.to.y].piece = NONE;

      h_position ^= h_values[move.color][ROOK][A][move.from.y];
      h_position ^= h_values[move.color][ROOK][D][move.from.y];
    }
  }

  /* Mise  jour du statut des roques si une case d'un des coins de l'chiquier
     est concerne : */
  if (   (move.from.x == A && move.from.y == 1)
      || (move.to.x == A && move.to.y == 1)) {

    /* Le grand roque blanc n'est plus possible : */
    CASTLE_STATUS &= 0x3D;
  }
  else if (   (move.from.x == A && move.from.y == 8)
           || (move.to.x == A && move.to.y == 8)) {

    /* Le grand roque noir n'est plus possible : */
    CASTLE_STATUS &= 0x37;
  }
  else if (   (move.from.x == H && move.from.y == 1)
           || (move.to.x == H && move.to.y == 1)) {

    /* Le petit roque blanc n'est plus possible : */
    CASTLE_STATUS &= 0x3E;
  }
  else if (   (move.from.x == H && move.from.y == 8)
           || (move.to.x == H && move.to.y == 8)) {

    /* Le petit roque noir n'est plus possible : */
    CASTLE_STATUS &= 0x3B;
  }

  /* Prise en compte des possibilits de roquer aprs le coup jou : */
  for (i = 1; i <= 4; i++) {

    h_position ^= h_values_castle[i][CASTLE_STATUS & castle_data_mask[i]];
  }

  /* Mise  jour du numro du coup de la partie : */
  if (active_color == BLACK) {

    fullmove++;
  }

  /* Le trait passe  la couleur oppose : */
  active_color ^= 1;
}
/* End of function : make */

/*******************************************************************************
  Function : unmake

  Description : Revient  la position prcdente sur l'chiquier.

  Validation level : 1 - moves.unmake
*******************************************************************************/
void unmake (move_s move) {

#ifdef DEBUG
debug.trace ("unmake", false, &unmake_calls);
#endif /* DEBUG */

  if (game_ply) {

    game_ply--;

#ifdef DEBUG_HIGH
  debug.print ("(%9d) - %20s", nodes, "unmake");
  debug.print (" : %d) %s%c%d%c%d%c\n",
    game_ply, active_color ? " ..... " : "",
    move.from.x + (int) 'a' - 1, move.from.y,
    move.to.x + (int) 'a' - 1, move.to.y,
    move.status == PROMOTION ? PIECE(move.promotion) : ' ');
#endif /* DEBUG_HIGH */

    h_position = h_position_history[game_ply];

    /* Mise  jour du comptage de demi-coups pour la rgle des 50 coups : */
    halfmove = halfmove_history[search_ply];

    /* Mise  jour des informations de la case de dpart : */
    board[move.from.x][move.from.y].piece = move.piece;
    board[move.from.x][move.from.y].color = move.color;
    board[move.from.x][move.from.y].form  = move.form;

    /* Mise  jour de la fiche associe : */
    pieces[move.color][move.form].x = move.from.x;
    pieces[move.color][move.form].y = move.from.y;

    /* Enlve l'indication de dplacement de la pice : */
    board[move.from.x][move.from.y].moved--;
    board[move.to.x][move.to.y].moved--;

    /* Effacement des informations de la case d'arrive : */
    if (move.captured_form == 0) {

      board[move.to.x][move.to.y].piece = NONE;
    }
    else {

      /* Si c'tait une prise en passant : */
      if (move.status == EN_PASSANT) {

        /* Le pion captur est remis sur l'chiquier : */
        board[move.to.x][move.to.y].piece   = NONE;
        board[move.to.x][move.from.y].piece = PAWN;
        board[move.to.x][move.from.y].color = move.color^1;
        board[move.to.x][move.from.y].form  = move.captured_form;
      }
      else {

        /* La pice capture est remise sur l'chiquier : */
        board[move.to.x][move.to.y].piece =
          pieces[move.color^1][move.captured_form].piece;
        board[move.to.x][move.to.y].color = move.color^1;
        board[move.to.x][move.to.y].form  = move.captured_form;

        /* Mise  jour du dcompte des pices pour dterminer la phase de la
           partie : */
        if (   board[move.to.x][move.to.y].piece > PAWN
            && board[move.to.x][move.to.y].piece < KING) {

          piece_count++;
        }
      }

      /* Et la fiche associe est mise  jour : */
      pieces[move.color^1][move.captured_form].captured = false;
    }

    /* Si la pice dplace tait un roi on mmorise sa nouvelle position : */
    if (move.piece == KING) {

      king[move.color].x = move.from.x;
      king[move.color].y = move.from.y;

      /* Si c'tait un petit roque la tour 'H' est dplace galement : */
      if (move.status == KING_CASTLE) {

        /* Mise  jour des informations de la case d'arrive : */
        board[H][move.to.y].piece = ROOK;
        board[H][move.to.y].color = move.color;
        board[H][move.to.y].form  = board[F][move.to.y].form;

        /* Indique que la tour n'a pas boug : */
        board[H][move.to.y].moved = 0;

        /* Mise  jour de la fiche associe : */
        pieces[move.color][board[H][move.to.y].form].x = H;

        /* Effacement des informations de la case de dpart : */
        board[F][move.to.y].piece = NONE;
      }

      /* Si c'tait un grand roque la tour 'A' est dplace galement : */
      if (move.status == QUEEN_CASTLE) {

        /* Mise  jour des informations de la case d'arrive : */
        board[A][move.to.y].piece = ROOK;
        board[A][move.to.y].color = move.color;
        board[A][move.to.y].form  = board[D][move.to.y].form;

        /* Indique que la tour n'a pas boug : */
        board[A][move.to.y].moved = 0;

        /* Mise  jour de la fiche associe : */
        pieces[move.color][board[A][move.to.y].form].x = A;

        /* Effacement des informations de la case de dpart : */
        board[D][move.to.y].piece = NONE;
      }
    }
    /* Si c'tait une promotion, la nature de la pice est change : */
    else if (move.status == PROMOTION) {

      board[move.from.x][move.from.y].piece = PAWN;

      pieces[move.color][move.form].piece = PAWN;
    }

    if (active_color == WHITE) {

      fullmove--;
    }

    active_color ^= 1;
  }
  else {

    tools.send_output (true, "   ERROR : There's no move to undo\n\n");
  }
}
/* End of function : unmake */

/*******************************************************************************
  Object declaration :
*******************************************************************************/

moves_module_s moves = {

  generate,
  is_attacked,
  make,
  unmake
};

/*******************************************************************************
  Function : search_auth_dir

  Description : Cette fonction permet de marquer les pices  mobilit rduite,
                et de prciser la direction (et celle oppose) qu'elles peuvent
                prendre en cas de clouage.

  Validation level : 1 - moves.search_auth_dir
*******************************************************************************/
void search_auth_dir (void) {

/* ------------------------------- Local data ------------------------------- */

bool look_forward;

int d;
int i;
int x, x0;
int y, y0;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("search_auth_dir\n", false, &search_auth_dir_calls);
#endif /* DEBUG */

  /* Par dfaut toutes les pices ont leurs directions autorises : */
  for (i = 1; i <= pieces_number[active_color]; i++) {

    pieces[active_color][i].authorized = -1;
  }

  for (d = 1; d <= 8; d++) {

    x = king[active_color].x;
    y = king[active_color].y;

    look_forward = true;
    while (look_forward) {

      x += vector[d].vx;
      y += vector[d].vy;

      if (IS_INSIDE(x, y)) {

        if (board[x][y].piece) {

          if (board[x][y].color == active_color) {

            x0 = x;
            y0 = y;

            while (look_forward) {

              x += vector[d].vx;
              y += vector[d].vy;

              if (IS_INSIDE(x, y)) {

                if (board[x][y].piece) {

                  look_forward = false;

                  if (board[x][y].color != active_color) {

                    if (liberty[board[x][y].piece].long_range == true) {

                      if (liberty[board[x][y].piece].direction[OPPOSED(d)]
                            == OPPOSED(d)) {

                        if (board[x0][y0].piece == PAWN) {

                          if (active_color == WHITE) {

                            if (d == 1 || d == 5) {

                              pieces[active_color]
                                    [board[x0][y0].form].authorized = 1;

                            } else if (d == 2 || d == 8) {

                              pieces[active_color]
                                    [board[x0][y0].form].authorized = d;

                            } else {

                              pieces[active_color]
                                    [board[x0][y0].form].authorized = 0;
                            }

                          } else {

                            if (d == 1 || d == 5) {

                              pieces[active_color]
                                    [board[x0][y0].form].authorized = 5;

                            } else if (d == 4 || d == 6) {

                              pieces[active_color]
                                    [board[x0][y0].form].authorized = d;

                            } else {

                              pieces[active_color]
                                    [board[x0][y0].form].authorized = 0;
                            }
                          }
                        }
                        else {

                          if (liberty[board[x0][y0].piece].direction[d] == d) {

                            pieces[active_color]
                                  [board[x0][y0].form].authorized = d;

                          } else {

                            pieces[active_color]
                                  [board[x0][y0].form].authorized = 0;
                          }
                        }
                      }
                    }
                  }
                }
              }
              else {
                look_forward = false;
              }
            }
          } else {
            look_forward = false;
          }
        }
      }
      else {
        look_forward = false;
      }
    }
  }
}
/* End of function : search_auth_dir */
