/*******************************************************************************
  File : search.c

  Description : Module de recherche.

  Ce module s'appuie principalement sur :
  1) l'algorithme min-max
  2) l'lagage alpha-beta
  3) le tri des coups gnrs pour acclrer l'lagage
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "book.h"
#include "debug.h"
#include "evaluate.h"
#include "hash.h"
#include "moves.h"
#include "search.h"
#include "tools.h"
#include "uci.h"

/*******************************************************************************
  Constants :
*******************************************************************************/

/* Permet principalement de donner une limite  la rflexion du programme dans
   le cas de la recherche d'une situation calme pour valuation : */
#define MAXIMUM_DEPTH        20

/* Limite maximum pour la recherche d'une situation calme : */
#define MAXIMUM_QUIET_DEPTH  20

/* Permet d'assurer au programme un minimum de rflexion : */
#define MINIMUM_DEPTH         5

/*******************************************************************************
  Types :
*******************************************************************************/

typedef struct {

  int score;

  move_s move1;
  move_s move2;

} killer_move_s;

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int think_calls = 0;
int first_search_calls = 0;
int minmax_search_calls = 0;
int pick_move_calls = 0;
int quiet_search_calls = 0;
int sort_moves_calls = 0;
#endif /* DEBUG */

#ifdef DEBUG_TREE
FILE *debug_tree_file;

struct {
  char type;
  int depth;
  int i;
  int moves_number;
  char lan_move[6];
  int alpha;
  int beta;
  int score;
} tree_nodes [DEBUG_TREE_LIMIT];

#endif /* DEBUG_TREE */

/* Pour privilgier la recherche de la variante principale : */
bool search_pv;

/* Mmorise l'interruption de la recherche sur manque de temps : */
bool time_exit;

/* Mmorise l'interruption de la recherche sur manque de temps : */
bool time_failure;

/* Mmorise le score atteint par le programme pendant la phase de recherche : */
int current_score;

/* Pour mmoriser les coups qui ont provoqu un lagage de l'arbre (rupture sur
   la borne alpha). Indic sur le nom de la pice, puis les cases de dpart et
   d'arrive : */
int heuristic[9][9][9][9];

/* Coups remarquables principalement  l'origine d'un lagage beta : */
killer_move_s killers[PV_LIMIT];

/*******************************************************************************
  Internal functions prototypes :
*******************************************************************************/

move_s first_search (int live_depth, int alpha, int beta);
int minmax_search (int live_depth, int alpha, int beta);
void sort_moves (move_s *legal_moves, int moves_number);
int quiet_search (int live_depth, int alpha, int beta);
int pick_move (move_s *legal_moves, int moves_number);

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : think

  Description : Initialisation de la rflexion du programme.

  Validation level : 1 - search.think
*******************************************************************************/
void think (move_s *move) {

/* ------------------------------- Local data ------------------------------- */

int elapsed_time;
int i;

move_s tmp_move;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("think\n", true, &think_calls);
#endif /* DEBUG */

  /* Le nombre de noeuds calcul est remis  zro : */
  nodes = 0;
  qnodes = 0;

  /* Recherche un coup dans la bibliothque des ouvertures : */
  if (!use_book || game_ply > BOOK_PLY_LIMIT || !book.select_move (move)) {

    /* Calcule le temps allou  la rflexion du programme : */
    tools.compute_time ();

    /* Efface la prcdente variante principale : */
    memset (pv, 0, sizeof (pv));

    /* Efface la prcdente heuristique : */
    memset (heuristic, 0, sizeof (heuristic));

    /* Efface les prcdents coups remarquables : */
    for (i = 0; i < PV_LIMIT; i++) {

      killers[i].score = -MIN_MAX_LIMIT;

      killers[i].move1 = null_move;
      killers[i].move2 = null_move;
    }

    /* Mmorise le nombre de pices prsentes sur l'chiquier au dbut de la
      recherche : */
    start_piece_count = piece_count;

    /* La recherche commence en augmentant progressivement la profondeur
       de rflexion, la limite se fera principalement sur le temps : */
    for (depth = 1; depth <= MAXIMUM_DEPTH; depth++) {

      /* Calcul du temps coul depuis le dbut de la rflexion : */
      elapsed_time = (int) difftime (time (0), start_time) * 1000;

      /* Si le temps pass est trop important, */
      if (   elapsed_time > time_for_move
          && depth > MINIMUM_DEPTH) {

        /* alors la rflexion du programme est interrompue. */
        break;
      }

      /* Lance la recherche  la racine de l'arbre : */
      tmp_move = first_search (depth, -MIN_MAX_LIMIT, MIN_MAX_LIMIT);

      /* A la fin de la rflexion, si au moins un coup est dtermin, */
      if (!time_failure) {

        /* alors prise en compte de ce dernier : */
        *move = tmp_move;

        if (depth >= MINIMUM_DEPTH) {

          /* L'tat de la rflexion est affich : */
          uci.info (current_score);
        }
      }

      /* Les scores des coup remarquables sont inutilisables pour les
         profondeurs suprieures, mais les coups peuvent servir pour
         l'algorithme de tri : */
      for (i = 0; i < PV_LIMIT; i++) {

        killers[i].score = -MIN_MAX_LIMIT;
      }
    }
  }

#ifdef DEBUG_TREE
{
int i, node;

  for (node = 1; node <= nodes && node < DEBUG_TREE_LIMIT; node++) {
    fprintf (debug_tree_file, "%6d %c%d ",
      node, tree_nodes[node].type, tree_nodes[node].depth);
    for (i = 0; i < tree_nodes[node].depth; i++) {
      fprintf (debug_tree_file, "   ");
    }
    fprintf (debug_tree_file, "%2d/%2d - %s = %4d (alpha %8d beta %8d)\n",
      tree_nodes[node].i, tree_nodes[node].moves_number, 
      tree_nodes[node].lan_move, tree_nodes[node].score,
      tree_nodes[node].alpha, tree_nodes[node].beta);
  }
  fprintf (debug_tree_file, "\nTotal analyzed nodes of tree : %d\n", nodes);
}
#endif /* DEBUG_TREE */
}
/* End of function : think */

/*******************************************************************************
  Object declaration :
*******************************************************************************/

search_module_s search = {

  think
};

/*******************************************************************************
  Internal functions :
*******************************************************************************/

/*******************************************************************************
  Function : first_search

  Description : Recherche au premier niveau dans l'arbre des positions.

  Validation level : 1 - search.first_search
*******************************************************************************/
move_s first_search (int live_depth, int alpha, int beta) {

/* ------------------------------- Local data ------------------------------- */

bool no_moves;

int i, j;
int moves_number;
int score = -MIN_MAX_LIMIT;
int search_extension = 0;

move_s best_move = null_move;
move_s legal_moves[MOVES_LIMIT];

#ifdef DEBUG_TREE
int debug_node;
#endif /* DEBUG_TREE */

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("first_search", true, &first_search_calls);
#endif /* DEBUG */

#ifdef DEBUG_HIGH
debug.print (" : live_depth %d alpha %d beta %d game_ply %d\n",
             live_depth, alpha, beta, game_ply);
#endif /* DEBUG_HIGH */

  search_ply = 1;

  /* Teste une ventuelle rptition de position : */
  if (tools.is_repeated ()) {

    current_score = 0;
    pv_length[search_ply] = 0;
    result = DRAW_REPETITION;

    return (null_move);
  }

  /* La longueur de la variante principale est rinitialise : */
  pv_length[search_ply] = 1;

  no_moves     = true;
  search_pv    = true;
  time_exit    = false;
  time_failure = false;

  /* Si le roi est attaqu, */
  if (moves.is_attacked (king[active_color].x, king[active_color].y,
                         active_color ^1)) {

    /* alors la recherche sera prolonge. */
    search_extension++;
  }

  /* Gnration de tous les coups lgaux de la position courante : */
  moves_number = moves.generate (legal_moves, LEGAL_MOVES);

  /* Tri des coups pour analyser en premier les plus intressants : */
  sort_moves (legal_moves, moves_number);

  /* Parcours de tous les coups lgaux dans l'ordre du tri prcdent : */
  while (i = pick_move (legal_moves, moves_number)) {

    /* Un nouveau noeud est atteint : */
    nodes++;

#ifdef DEBUG_HIGH
  debug.new_debug_files ();
#endif /* DEBUG_HIGH */

#ifdef DEBUG_TREE
if (nodes < DEBUG_TREE_LIMIT) {
  tools.convert_move (legal_moves[i], tree_nodes[nodes].lan_move);

  tree_nodes[nodes].type         = 'f';
  tree_nodes[nodes].depth        = 1;
  tree_nodes[nodes].i            = i;
  tree_nodes[nodes].moves_number = moves_number;
  tree_nodes[nodes].alpha        = alpha;
  tree_nodes[nodes].beta         = beta;

  debug_node = nodes;
}
#endif /* DEBUG_TREE */

    /* Poursuite de l'analyse au niveau suivant de l'arbre : */
    moves.make (legal_moves[i]);
    search_ply++;
    score = -minmax_search (live_depth - 1 + search_extension, -beta, -alpha);
    search_ply--;
    moves.unmake (legal_moves[i]);

    /* Si la rflexion a t interrompue lors du parcours de l'arbre, */
    if (time_exit) {

      if (no_moves) {

        time_failure = true;
      }

      return best_move;
    }

    no_moves = false;

    /* Si l'valuation est suprieure  la prcdente, */
    if (score > alpha) {

      /* alors la nouvelle valuation devient la rfrence, */
      alpha = score;
      current_score = score;

      /* et le coup analys devient le meilleur pour le programme. */
      best_move = legal_moves[i];

      /* L'intrt du coup trouv augmente pour un prochain tri : */
      heuristic[legal_moves[i].from.x][legal_moves[i].from.y]
               [legal_moves[i].to.x][legal_moves[i].to.y] += live_depth;

      /* Vrifie si le coup fait partie des deux meilleurs coups : */
      if (score > killers[search_ply].score) {

        killers[search_ply].score = score;
        killers[search_ply].move2 = killers[search_ply].move1;
        killers[search_ply].move1 = legal_moves[i];
#ifdef DEBUG_HIGH
debug.print ("(%9d) - killer : score %d - move1 %c%d%c%d%c",
  nodes, score,
  legal_moves[i].from.x + (int) 'a' - 1, legal_moves[i].from.y,
  legal_moves[i].to.x + (int) 'a' - 1, legal_moves[i].to.y,
  legal_moves[i].status == PROMOTION ? PIECE(legal_moves[i].promotion) : ' ');
if (killers[search_ply].move2.piece) {
debug.print (" - move2 %c%d%c%d%c\n",
  killers[search_ply].move2.from.x + (int) 'a' - 1,
  killers[search_ply].move2.from.y,
  killers[search_ply].move2.to.x + (int) 'a' - 1,
  killers[search_ply].move2.to.y,
  killers[search_ply].move2.status == PROMOTION ?
    PIECE(killers[search_ply].move2.promotion) : ' ');
}
else {
debug.print ("\n");
}
#endif /* DEBUG_HIGH */
      }

      /* Mise  jour de la variante principale : */
      pv[search_ply][search_ply] = legal_moves[i];
      for (j = search_ply + 1; j < pv_length[search_ply + 1]; j++) {

        pv[search_ply][j] = pv[search_ply + 1][j];
      }
      pv_length[search_ply] = pv_length[search_ply + 1];

      if (depth >= MINIMUM_DEPTH) {

        /* L'tat de la rflexion est affich : */
        uci.info (alpha);
      }
    }
#ifdef DEBUG_TREE
if (nodes < DEBUG_TREE_LIMIT) {
  tree_nodes[debug_node].score = score;
}
#endif /* DEBUG_TREE */
  }

  /* Si la liste des coups lgaux est vide, */
  if (moves_number == 0) {

    /* et si le roi est attaqu par la couleur adverse, */
    if (moves.is_attacked (king[active_color].x, king[active_color].y,
                           active_color^1)) {

      /* alors il y a chec et mat. */
      result = active_color ? WHITE_IS_MATED : BLACK_IS_MATED;
    }
    else {

      /* sinon il y a pat. */
      result == DRAW_STALEMATE;
    }
  }
  else {

    /* Teste si la rgle des 50 coups s'applique : */
    if (halfmove > 100) {

      current_score = 0;
      pv_length[search_ply] = 0;
      result = DRAW_FIFTY_MOVES;

      return (null_move);
    }
  }

  /* Cherche si l'adversaire du programme est mat : */
  if (score == MIN_MAX_LIMIT - 2) {

    result = active_color ? BLACK_IS_MATED : WHITE_IS_MATED;
  }

  /* Le programme remonte le meilleur coup trouv. */
  return best_move;
}
/* End of function : first_search */

/*******************************************************************************
  Function : minmax_search

  Description : Cette fonction est appele rcursivement pour parcourir l'arbre
                des positions et remonter l'valuation de chacune  la racine
                Algorithme min-max avec une recherche des situations calmes

  Validation level : 1 - search.minmax_search
*******************************************************************************/
int minmax_search (int live_depth, int alpha, int beta) {

/* ------------------------------- Local data ------------------------------- */

int elapsed_time;
int i, j;
int moves_number;
int score;
int search_extension = 0;

move_s legal_moves[MOVES_LIMIT];

#ifdef DEBUG_TREE
int debug_node;
#endif /* DEBUG_TREE */

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("minmax_search", true, &minmax_search_calls);
#endif /* DEBUG */

#ifdef DEBUG_HIGH
debug.print (" : live_depth %d alpha %d beta %d nodes %d\n",
             live_depth, alpha, beta, nodes);
#endif /* DEBUG_HIGH */

  if (depth >= MINIMUM_DEPTH) {

    elapsed_time = (int) difftime (time (0), start_time) * 1000;

    if (elapsed_time > time_for_move) {

      time_exit = true;
      return 0;
    }
  }

  /* Teste une ventuelle rptition de position : */
  if (tools.is_repeated ()) {

    return 0;
  }

  /* A chaque nouvel appel de la fonction, la longueur de la variante
     principale augmente : */
  pv_length[search_ply] = search_ply;

  /* Si le roi est attaqu, */
  if (moves.is_attacked (king[active_color].x, king[active_color].y,
                         active_color ^1)) {

    /* tout en vitant l'chec perptuel, */
    if (search_ply < MAXIMUM_DEPTH) {

      /* alors la recherche sera prolonge. */
      search_extension++;
    }
  }

  /* Si le niveau limite est atteint : */
  if (!(live_depth + search_extension)) {

    /* alors le programme recherche une situation calme : */
    score = quiet_search (MAXIMUM_QUIET_DEPTH, alpha, beta);

    /* avant de remonter l'valuation de la position obtenue : */
    return score;
  }

  /* Gnration de la liste des coups lgaux dans la position courante : */
  moves_number = moves.generate (legal_moves, LEGAL_MOVES);

  /* Tri des coups pour analyser en premier les plus intressants : */
  sort_moves (legal_moves, moves_number);

  /* Parcours de tous les coups lgaux dans l'ordre du tri prcdent : */
  while (i = pick_move (legal_moves, moves_number)) {

    /* Un nouveau noeud est atteint : */
    nodes++;

#ifdef DEBUG_HIGH
  debug.new_debug_files ();
#endif /* DEBUG_HIGH */

#ifdef DEBUG_TREE
if (nodes < DEBUG_TREE_LIMIT) {
  tools.convert_move (legal_moves[i], tree_nodes[nodes].lan_move);

  tree_nodes[nodes].type         = 'n';
  tree_nodes[nodes].depth        = depth - live_depth + 1;
  tree_nodes[nodes].i            = i;
  tree_nodes[nodes].moves_number = moves_number;
  tree_nodes[nodes].alpha        = alpha;
  tree_nodes[nodes].beta         = beta;

  debug_node = nodes;
}
#endif /* DEBUG_TREE */

    /* Poursuite de l'analyse au niveau suivant de l'arbre : */
    moves.make (legal_moves[i]);
    search_ply++;
    score = -minmax_search (live_depth - 1 + search_extension, -beta, -alpha);
    search_ply--;
    moves.unmake (legal_moves[i]);

#ifdef DEBUG_TREE
if (nodes < DEBUG_TREE_LIMIT) {
  tree_nodes[debug_node].score = score;
}
#endif /* DEBUG_TREE */

    /* Si la rflexion a t interrompue lors du parcours de l'arbre, */
    if (time_exit) {

      /* alors on propage cette interruption. */
      return 0;
    }

    /* Si l'valuation est suprieure  la prcdente, */
    if (score > alpha) {

      /* alors le coup analys est intressant pour un futur tri, */
      heuristic[legal_moves[i].from.x][legal_moves[i].from.y]
               [legal_moves[i].to.x][legal_moves[i].to.y] += live_depth;

      /* et si il y a une coupure beta : */
      if (score >= beta) {

        /* Vrifie si le coup fait partie des deux meilleurs coups : */
        if (score > killers[search_ply].score) {

          killers[search_ply].score = score;
          killers[search_ply].move2 = killers[search_ply].move1;
          killers[search_ply].move1 = legal_moves[i];
#ifdef DEBUG_HIGH
debug.print ("(%9d) - killer : score %d - move1 %c%d%c%d%c",
  nodes, score,
  legal_moves[i].from.x + (int) 'a' - 1, legal_moves[i].from.y,
  legal_moves[i].to.x + (int) 'a' - 1, legal_moves[i].to.y,
  legal_moves[i].status == PROMOTION ? PIECE(legal_moves[i].promotion) : ' ');
if (killers[search_ply].move2.piece) {
debug.print (" - move2 %c%d%c%d%c\n",
  killers[search_ply].move2.from.x + (int) 'a' - 1,
  killers[search_ply].move2.from.y,
  killers[search_ply].move2.to.x + (int) 'a' - 1,
  killers[search_ply].move2.to.y,
  killers[search_ply].move2.status == PROMOTION ?
    PIECE(killers[search_ply].move2.promotion) : ' ');
}
else {
debug.print ("\n");
}
#endif /* DEBUG_HIGH */
        }

        /* et rien ne sert de continuer  analyser les autres coups situs
           au mme niveau de profondeur (lagage beta) : */
        return beta;
      }

      /* la nouvelle valuation devient la rfrence : */
      alpha = score;

      /* mise  jour de la variante principale : */
      pv[search_ply][search_ply] = legal_moves[i];
      for (j = search_ply + 1; j < pv_length[search_ply + 1]; j++) {

        pv[search_ply][j] = pv[search_ply + 1][j];
      }
      pv_length[search_ply] = pv_length[search_ply + 1];
    }
  }

  /* Si aucun coup lgal n'existe dans la position, */
  if (moves_number == 0) {

    /* et si le roi est en chec : */
    if (moves.is_attacked (king[active_color].x, king[active_color].y,
                           active_color^1)) {

      /* alors un mat est trouv (ici, search_ply indique le nombre de
         demi-coups avant le mat) : */
      alpha = - MIN_MAX_LIMIT + search_ply;
    }
    else {

      /* sinon il y a pat, la partie sera nulle. */
      alpha = 0;
    }
  }
  else {

    /* Teste si la rgle des 50 coups s'applique : */
    if (halfmove > 100) {

      return (0);
    }
  }

  return alpha;
}
/* End of function : minmax_search */

/*******************************************************************************
  Function : pick_move

  Description : A chaque appel, retourne le meilleur coup  analyser.

  Validation level : 1 - search.pick_move
*******************************************************************************/
int pick_move (move_s *legal_moves, int moves_number) {

/* ------------------------------- Local data ------------------------------- */

int best_note = -1;
int i;
int move = 0; /* Le coup 0 n'existe pas, il indiquera l'absence de rsultat. */

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("pick_move", true, &pick_move_calls);
#endif /* DEBUG */

  for (i = 1; i <= moves_number; i++) {

    if (legal_moves[i].note > best_note) {

      best_note = legal_moves[i].note;
      move = i;
    }
  }

  legal_moves[move].note = -1;

#ifdef DEBUG_HIGH
debug.print (" : move %d\n", move);
#endif /* DEBUG_HIGH */
  return move;
}
/* End of function : pick_move */

/*******************************************************************************
  Function : quiet_search

  Description : Recherche d'une position calme qui permette une valuation.

  Validation level : 1 - search.quiet_search
*******************************************************************************/
int quiet_search (int live_depth, int alpha, int beta) {

/* ------------------------------- Local data ------------------------------- */

int i, j;
int initial_score;
int moves_number;
int score;

move_s legal_moves[MOVES_LIMIT];

#ifdef DEBUG_TREE
int debug_node;
#endif /* DEBUG_TREE */

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("quiet_search", true, &quiet_search_calls);
#endif /* DEBUG */

#ifdef DEBUG_HIGH
debug.print (" : live_depth %d alpha %d beta %d nodes %d\n",
             live_depth, alpha, beta, nodes);
#endif /* DEBUG_HIGH */

  /* Evaluation de la position courante  l'appel de la fonction : */
  initial_score = evaluate.evaluate_position (false);

  /* Si la profondeur maximum est atteinte ou si il y a une coupure beta : */
  if (!live_depth || initial_score >= beta) {

    /* alors on remonte directement l'valuation obtenue. */
    return initial_score;
  }

  /* La longueur de la variante principale augmente : */
  pv_length[search_ply] = search_ply;

  /* Si l'valuation est suprieure  la prcdente, */
  if (initial_score > alpha) {

    /* alors la nouvelle valuation devient la rfrence : */
    alpha = initial_score;
  }

  /* Gnration des seules captures possibles dans la position courante : */
  moves_number = moves.generate (legal_moves, CAPTURES_ONLY);

  /* Tri des coups pour analyser en premier les plus intressants : */
  sort_moves (legal_moves, moves_number);

  /* Parcours de toutes les captures dans l'ordre du tri prcdent : */
  while (i = pick_move (legal_moves, moves_number)) {

    /* Un nouveau noeud est atteint : */
    nodes++;
    qnodes++;

#ifdef DEBUG_HIGH
  debug.new_debug_files ();
#endif /* DEBUG_HIGH */

#ifdef DEBUG_TREE
if (nodes < DEBUG_TREE_LIMIT) {
  tools.convert_move (legal_moves[i], tree_nodes[nodes].lan_move);

  tree_nodes[nodes].type         = 'q';
  tree_nodes[nodes].depth        = MAXIMUM_QUIET_DEPTH - live_depth + 1;
  tree_nodes[nodes].i            = i;
  tree_nodes[nodes].moves_number = moves_number;
  tree_nodes[nodes].alpha        = alpha;
  tree_nodes[nodes].beta         = beta;

  debug_node = nodes;
}
#endif /* DEBUG_TREE */

    /* Poursuite de l'analyse au niveau suivant de l'arbre : */
    moves.make (legal_moves[i]);
    search_ply++;
    score = - quiet_search (live_depth - 1, - beta, - alpha);
    search_ply--;
    moves.unmake (legal_moves[i]);

#ifdef DEBUG_TREE
if (nodes < DEBUG_TREE_LIMIT) {
  tree_nodes[debug_node].score = score;
}
#endif /* DEBUG_TREE */

    /* Si l'valuation est suprieure  la prcdente, */
    if (score > alpha) {

      /* et si il y a une coupure beta, */
      if (score >= beta) {

        /* alors on vrifie si le coup fait partie des deux meilleurs coups : */
        if (score > killers[search_ply].score) {

          killers[search_ply].score = score;
          killers[search_ply].move2 = killers[search_ply].move1;
          killers[search_ply].move1 = legal_moves[i];
#ifdef DEBUG_HIGH
debug.print ("(%9d) - killer : score %d - move1 %c%d%c%d%c",
  nodes, score,
  legal_moves[i].from.x + (int) 'a' - 1, legal_moves[i].from.y,
  legal_moves[i].to.x + (int) 'a' - 1, legal_moves[i].to.y,
  legal_moves[i].status == PROMOTION ? PIECE(legal_moves[i].promotion) : ' ');
if (killers[search_ply].move2.piece) {
debug.print (" - move2 %c%d%c%d%c\n",
  killers[search_ply].move2.from.x + (int) 'a' - 1,
  killers[search_ply].move2.from.y,
  killers[search_ply].move2.to.x + (int) 'a' - 1,
  killers[search_ply].move2.to.y,
  killers[search_ply].move2.status == PROMOTION ?
    PIECE(killers[search_ply].move2.promotion) : ' ');
}
else {
debug.print ("\n");
}
#endif /* DEBUG_HIGH */
        }

        /* et rien ne sert de continuer  analyser les autres coups situs
           au mme niveau de profondeur (lagage beta) : */
        return beta;
      }

      /* la nouvelle valuation devient la rfrence : */
      alpha = score;

      /* Mise  jour de la variante principale : */
      pv[search_ply][search_ply] = legal_moves[i];
      for (j = search_ply + 1; j < pv_length[search_ply + 1]; j++) {

        pv[search_ply][j] = pv[search_ply + 1][j];
      }
      pv_length[search_ply] = pv_length[search_ply + 1];
    }
  }

  /* Si aucune capture n'existe  ce niveau de profondeur, */
  if (moves_number == 0) {

    /* alors on remonte l'valuation de la position courante. */
    return initial_score;
  }
  else {

    /* sinon on remonte le meilleur score de l'arbre : */
    return alpha;
  }
}
/* End of function : quiet_search */

/*******************************************************************************
  Function : sort_moves

  Description : Pour lancer l'analyse en priorit sur les coups les plus
                intressants, ceci afin de provoquer un lagage au plus tt.

  Validation level : 1 - search.sort_moves
*******************************************************************************/
void sort_moves (move_s *legal_moves, int moves_number) {

/* ------------------------------- Local data ------------------------------- */

int i;

#define X0 legal_moves[i].from.x
#define Y0 legal_moves[i].from.y
#define X1 legal_moves[i].to.x
#define Y1 legal_moves[i].to.y
#define PROMOTED legal_moves[i].promotion
#define CAPTURED pieces[legal_moves[i].color^1] \
                       [legal_moves[i].captured_form].piece

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("sort_moves", true, &sort_moves_calls);
#endif /* DEBUG */

  /* Pendant la recherche de la variante principale, les coups correspondants
     auront une note plus importante : */
  if (search_pv) {

    search_pv = false;

    for (i = 1; i <= moves_number; i++)   {

      /* Tri en fonction du gain matriel : */
      if (CAPTURED) {

        legal_moves[i].note =
          pieces_value[CAPTURED] - pieces_value[board[X0][Y0].piece] + 1000;
#ifdef DEBUG_HIGH
debug.print (" (1 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */
      }
      else {

        legal_moves[i].note = 0;
      }

      /* Donne la prpondrance  la variante principale. De plus les coups sont
         dj tris : */
      if (   X0 == pv[1][search_ply].from.x
          && Y0 == pv[1][search_ply].from.y
          && X1 == pv[1][search_ply].to.x
          && Y1 == pv[1][search_ply].to.y
          && PROMOTED == pv[1][search_ply].promotion) {

        search_pv = true;
        legal_moves[i].note += MIN_MAX_LIMIT;
#ifdef DEBUG_HIGH
debug.print (" (2 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */
      }
      else {

        /* Ajoute le bonus li  l'intrt de ce coup dans les prcdentes
           recherches  : */
        legal_moves[i].note += heuristic[X0][Y0][X1][Y1];
#ifdef DEBUG_HIGH
debug.print (" (3 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */

        /* Ajoute le bonus li aux coups remarquables : */
        if (   X0 == killers[search_ply].move1.from.x
            && Y0 == killers[search_ply].move1.from.y
            && X1 == killers[search_ply].move1.to.x
            && Y1 == killers[search_ply].move1.to.y
            && PROMOTED == killers[search_ply].move1.promotion) {

          legal_moves[i].note += 1000;
#ifdef DEBUG_HIGH
debug.print (" (4 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */
        }
        else if (   X0 == killers[search_ply].move2.from.x
                 && Y0 == killers[search_ply].move2.from.y
                 && X1 == killers[search_ply].move2.to.x
                 && Y1 == killers[search_ply].move2.to.y
                 && PROMOTED == killers[search_ply].move2.promotion) {

          legal_moves[i].note += 500;
#ifdef DEBUG_HIGH
debug.print (" (5 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */
        }
      }
    }
  }
  else {

    for (i = 1; i <= moves_number; i++) {

      /* Tri en fonction du gain matriel : */
      if (CAPTURED) {

        legal_moves[i].note =
          pieces_value[CAPTURED] - pieces_value[board[X0][Y0].piece] + 1000;
#ifdef DEBUG_HIGH
debug.print (" (6 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */
      }
      else {

        legal_moves[i].note = 0;
      }

      /* Ajoute le bonus li  l'intrt de ce coup dans les prcdentes
         recherches  : */
      legal_moves[i].note += heuristic[X0][Y0][X1][Y1];

#ifdef DEBUG_HIGH
debug.print (" (7 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */

      /* Ajoute le bonus li aux coups remarquables : */
      if (   X0 == killers[search_ply].move1.from.x
          && Y0 == killers[search_ply].move1.from.y
          && X1 == killers[search_ply].move1.to.x
          && Y1 == killers[search_ply].move1.to.y
          && PROMOTED == killers[search_ply].move1.promotion) {

        legal_moves[i].note += 1000;
#ifdef DEBUG_HIGH
debug.print (" (8 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */
      }
      else if (   X0 == killers[search_ply].move2.from.x
               && Y0 == killers[search_ply].move2.from.y
               && X1 == killers[search_ply].move2.to.x
               && Y1 == killers[search_ply].move2.to.y
               && PROMOTED == killers[search_ply].move2.promotion) {

        legal_moves[i].note += 500;
#ifdef DEBUG_HIGH
debug.print (" (9 i%d %d)", i, legal_moves[i].note);
#endif /* DEBUG_HIGH */
      }
    }
  }

#ifdef DEBUG_HIGH
debug.print (" : ");
for (i = 1; i <= moves_number; i++) {
  debug.print ("%c%d%c%d%c %d - ",
    legal_moves[i].from.x + (int) 'a' -1, legal_moves[i].from.y,
    legal_moves[i].to.x + (int) 'a' -1, legal_moves[i].to.y,
    legal_moves[i].status == PROMOTION ? PIECE(legal_moves[i].promotion) : ' ',
    legal_moves[i].note);
}
debug.print ("\n");
#endif /* DEBUG_HIGH */

#undef X0
#undef Y0
#undef X1
#undef Y1
#undef PROMOTED
#undef CAPTURED
}
/* End of function : sort_moves */
