/*******************************************************************************
  File : tools.c

  Description : Diverses fonctions utilises tout au long du programme.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "debug.h"
#include "evaluate.h"
#include "data.h"
#include "hash.h"
#include "moves.h"
#include "tools.h"
#include "uci.h"

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int compute_time_calls = 0;
int convert_lan_move_calls = 0;
int convert_move_calls = 0;
int generate_random_calls = 0;
int is_repeated_calls = 0;
int print_board_calls = 0;
int read_input_calls = 0;
int send_output_calls = 0;
int update_pieces_calls = 0;
int verify_move_calls = 0;
#endif /* DEBUG */

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : compute_time

  Description : Avant chaque coup du programme, cette fonction calcule le temps
                octroy au programme avant d'arrter sa rflexion.
                Les temps sont calculs en millimes de secondes.

  Validation level : 1 - tools.compute_time
*******************************************************************************/
void compute_time (void) {

/* ------------------------------- Local data ------------------------------- */

bool time_is_minimum = false;
bool time_is_maximum = false;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("compute_time", true, &compute_time_calls);
#endif /* DEBUG */

  /* En mode UCI, le programme calcule un temps par coup en fonction du temps
     allou pour la partie : */
  if (uci_mode) {

    /* Pour la premire fois, */
    if (base_time == 0) {

      /* le temps de base allou est arbitrairement calcul sur une base de 60
         coups pour la partie : */
      base_time = (computer_color ? uci_wtime : uci_btime) / 60.0;
      time_for_move = base_time;
    }

    /* Si le programme est en avance au temps par rapport  son adversaire, */
    if ((computer_color ? uci_wtime : uci_btime) >
        (computer_color ? uci_btime : uci_wtime)) {

      /* alors le temps allou pour le prochain coup est augment de 10% : */
      time_for_move = time_for_move * 1.1;

        /* Pour viter un emballement, une limite maximum est fixe : */
      if (time_for_move > 2 * base_time) {

        time_for_move = 2 * base_time;
        time_is_maximum = true;
      }
    }
    else {

      /* le temps allou est diminu de 20% : */
        time_for_move = time_for_move * 0.8;

       /* Pour permettre une certaine pertinence de l'analyse, une limite
          minimum est fixe : */
       if (time_for_move < base_time / 2.5) {

         time_for_move = base_time / 2.5;
         time_is_minimum = true;
       }
    }

    /* S'il reste moins de trente secondes, */
    if ((computer_color ? uci_wtime : uci_btime) < 30000) {

       /* un temps maximum est octroy pour viter le zeitnot (la profondeur
          minimum de rflexion vitera les cas limites) : */
       if (time_for_move > (computer_color ? uci_wtime : uci_btime) / 10.0) {

         time_for_move = (computer_color ? uci_wtime : uci_btime) / 10.0;
         time_is_maximum = true;
       }
    }

    printf ("info string Time for move : %.1f s (base time %.1f%s)\n",
      time_for_move / 1000.0, base_time / 1000.0,
      time_is_maximum ? " - time is limited" :
      time_is_minimum ? " - minimum time allowed" : "");
  }
  else {

    /* en mode console le temps pour chaque coup est fixe (10 secondes) : */
    time_for_move = 10000;
  }

#ifdef DEBUG_HIGH
debug.print (" : time_for_move %d\n", time_for_move);
#endif /* DEBUG_HIGH */
}
/* End of function : compute_time */

/*******************************************************************************
  Function : convert_lan_move

  Description : Pour convertir un coup au format LAN, dans un format connu par
                le programme (voir la structure "move_s").

  Validation level : 1 - tools.convert_lan_move
*******************************************************************************/
bool convert_lan_move (const char *lan_move, move_s *move) {

/* ------------------------------- Local data ------------------------------- */

bool is_legal = false;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("convert_lan_move\n", false, &convert_lan_move_calls);
#endif /* DEBUG */

  /* Initialisation de la nouvelle fiche : */
  memset (move, 0, sizeof (move_s));

  /* Calcul des cases de dpart et d'arrive : */
  if (*lan_move >= 'a' && *lan_move <= 'h') {

    move->from.x = (*lan_move) - 'a' + 1;

    lan_move++;
    if (*lan_move >= '1' && *lan_move <= '8') {

      move->from.y = (*lan_move) - '1' + 1;

      lan_move++;
      if (*lan_move >= 'a' && *lan_move <= 'h') {

        move->to.x = (*lan_move) - 'a' + 1;

        lan_move++;
        if (*lan_move >= '1' && *lan_move <= '8') {

          /* Le coup est au format LAN : */
          is_legal = true;

          move->to.y = (*lan_move) - '1' + 1;
        }
      }
    }
  }

  /* Recherche d'une ventuelle promotion de pion : */
  lan_move++;
  switch (*lan_move) {
    case 'n' : move->status = PROMOTION; move->promotion = KNIGHT; break;
    case 'b' : move->status = PROMOTION; move->promotion = BISHOP; break;
    case 'r' : move->status = PROMOTION; move->promotion = ROOK;   break;
    case 'q' : move->status = PROMOTION; move->promotion = QUEEN;  break;
  }

  /* Vrifie la prsence d'une pice sur la case de dpart : */
  if (board[move->from.x][move->from.y].piece) {

    /* Met  jour le nom de la pice et sa couleur : */
    move->piece = board[move->from.x][move->from.y].piece;
    move->color = board[move->from.x][move->from.y].color;
    move->form  = board[move->from.x][move->from.y].form;

    /* Vrifie si le coup est un roque : */
    if (   move->piece == KING
        && DISTANCE(move->from.x, move->from.y, move->to.x, move->to.y) == 2) {

      if (move->to.x == G) {

        move->status = KING_CASTLE;
      }
      else if (move->to.x == C) {

        move->status = QUEEN_CASTLE;
      }
      else {

        is_legal = false;
      }
    }

    /* Vrifie si le coup permet la capture d'une pice : */
    if (board[move->to.x][move->to.y].piece) {

      move->captured_form = board[move->to.x][move->to.y].form;
    }

    /* Vrifie si le coup est une prise en passant : */
    if (   board[move->from.x][move->from.y].piece == PAWN
        && !board[move->to.x][move->to.y].piece
        && move->from.x != move->to.x) {

      move->status = EN_PASSANT;
      move->captured_form = board[move->to.x][move->from.y].form;
    }
  }
  else {

    is_legal = false;
  }

  return is_legal;
}
/* End of function : convert_lan_move */

/*******************************************************************************
  Function : convert_move

  Description : Pour convertir un coup tel qu'il est connu du programme, dans la
                notation algbrique standard (LAN : Long Algebraic Notation).

  Validation level : 1 - tools.convert_move
*******************************************************************************/
void convert_move (move_s move, char lan_move[]) {

/* ------------------------------- Local data ------------------------------- */

char promoted_char;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("convert_move\n", false, &convert_move_calls);
#endif /* DEBUG */

  switch (move.promotion) {
    case (KNIGHT) : promoted_char = 'n'; break;
    case (BISHOP) : promoted_char = 'b'; break;
    case (ROOK)   : promoted_char = 'r'; break;
    case (QUEEN)  : promoted_char = 'q'; break;
    default :
      promoted_char = '\0';
  }

  sprintf (lan_move, "%c%d%c%d%c",
           move.from.x + (int) 'a' - 1, move.from.y,
           move.to.x + (int) 'a' - 1, move.to.y,
           promoted_char);
}
/* End of function : convert_move */

/*******************************************************************************
  Function : generate_random

  Description : Renvoie un nombre alatoire sur 32 bits.

  Validation level : 1 - tools.generate_random
*******************************************************************************/
unsigned int generate_random (void) {

/* ------------------------------- Local data ------------------------------- */

int byte1, byte2, byte3, byte4;
int random_number;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("generate_random\n", false, &generate_random_calls);
#endif /* DEBUG */

  /* Gnration de 4 nombres compris dans l'intervalle [0;256[ avec une
     prcision de 1/RAND_MAX */
  byte1 = (int) ((double) rand() * 256 / (RAND_MAX + 1));
  byte2 = (int) ((double) rand() * 256 / (RAND_MAX + 1));
  byte3 = (int) ((double) rand() * 256 / (RAND_MAX + 1));
  byte4 = (int) ((double) rand() * 256 / (RAND_MAX + 1));

  /* Concatnation des 4 octets pour former un nombre de 32 bits : */
  random_number = byte1 + (byte2 << 8) + (byte3 << 16) + (byte4 << 24);

  return random_number;
}
/* End of function : generate_random */

/*******************************************************************************
  Function : is_repeated

  Description : Recherche les rptitions de position.

  Validation level : 1 - tools.is_repeated
*******************************************************************************/
bool is_repeated (void) {

/* ------------------------------- Local data ------------------------------- */

int count = 1;
int i;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("is_repeated", true, &is_repeated_calls);
#endif /* DEBUG */

  if (halfmove >= 6) {

    for (i = game_ply - 2; i >= game_ply - halfmove; i -= 2) {

      if (h_position == h_position_history[i]) {

        count++;
      }

      if (count > 2) {

#ifdef DEBUG_HIGH
debug.print (" : true\n");
#endif /* DEBUG_HIGH */
        return true;
      }
    }
  }

#ifdef DEBUG_HIGH
debug.print (" : false\n");
#endif /* DEBUG_HIGH */
  return false;
}
/* End of function : is_repeated */

/*******************************************************************************
  Function : print_board

  Description : Enregistre une reprsentation de l'chiquier dans le fichier
                donn en entre (console ou autre).

  Validation level : 1 - tools.print_board
*******************************************************************************/
void print_board (FILE *output_file) {

/* ------------------------------- Local data ------------------------------- */

char *line = "   |---+---+---+---+---+---+---+---|";
char *piece_char[2][7] = {
  { " ", "p", "n", "b", "r", "q", "k" },
  { " ", "P", "N", "B", "R", "Q", "K" }
};

int x, y;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("print_board\n", false, &print_board_calls);
#endif /* DEBUG */

  fprintf (output_file, "  %s\n", line);

  for (y = 8; y > 0; y--) {

    fprintf (output_file, "   %d |", y);

    for (x = 1; x < 9; x++) {

      fprintf (output_file, " %s |",
               piece_char[board[x][y].color][board[x][y].piece]);
    }

#if defined DEBUG && defined NO_DEBUG_HIGH
    if (y == 7) {

      fprintf (output_file, " %.2f (static evaluation)",
               (float) evaluate.evaluate_position (true) / 100);
    }
    if (y == 6) {

      if  (tools.is_repeated ()) {

        fprintf (output_file, " Game is draw");
      }
      else {

        fprintf (output_file, " %d (halfmove)", halfmove);
      }
    }
    if (y == 4) {

      fprintf (output_file, " %x (hash)", h_position);
    }
    if (y == 3) {

      fprintf (output_file, " %x (castle_status)", CASTLE_STATUS);
    }
#endif /* DEBUG */

    if (y == 2) {

      fprintf (output_file, " %d (game_ply)", game_ply);
    }

    if (y == 1) {

      fprintf (output_file, active_color ? " WHITE to move" : " BLACK to move");
    }

    fprintf (output_file, "\n  %s\n", line);
  }

  fprintf (output_file, "       a   b   c   d   e   f   g   h\n\n");
}
/* End of function : print_board */

/*******************************************************************************
  Function : read_input

  Description : Lecture de caractres  partir de l'entre standard.

  Validation level : 1 - tools.read_input
*******************************************************************************/
void read_input (char string[]) {

/* ------------------------------- Local data ------------------------------- */

int character;
int i = 0;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("read_input", true, &read_input_calls);
#endif /* DEBUG */

/*  debug.close_debug_files (); */

  /* Lecture d'une suite de caractres : */
  while ((character = getc (stdin)) != (int) '\n' && character != EOF) {

    if (i < STRING_LENGTH - 1) {

      string[i++] = character;
    }
  }

  /* Pour terminer la chane : */
  string[i] = '\0';

/*  debug.open_debug_files (); */

  /* Pour tracer la saisie dans le fichier journal : */
  tools.send_output (false, "\nInput  : %s\n", string);

#ifdef DEBUG_HIGH
debug.print (" : %s\n", string);
#endif /* DEBUG_HIGH */
}
/* End of function : read_input */

/*******************************************************************************
  Function : send_output

  Description : Pour envoyer des informations vers les diffrentes sorties
                (console, fichier journal).

  Validation level : 1 - tools.send_output
*******************************************************************************/
int send_output (bool to_stdout, const char *format, ...) {

/* ------------------------------- Local data ------------------------------- */

int size;

va_list list;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("send_output", false, &send_output_calls);
#endif /* DEBUG */

  /* Initialisation de la liste des paramtres : */
  va_start (list, format);

  if (to_stdout) {

    /* Envoi des donnes vers la console : */
    size = vprintf (format, list);
  }

#ifdef DEBUG
  /* Ecriture des donnes dans le fichier journal : */
  size = vfprintf (log_file, format, list);
  fflush (log_file);
#endif /* DEBUG */

  /* Fin de lecture de la liste des paramtres : */
  va_end (list);

  return size;
}
/* End of function : send_output */

/*******************************************************************************
  Function : update_pieces

  Description : Met  jour la table des pices et le lien entre la fiche de la
                pice et la case de l'chiquier sur laquelle elle se trouve.
                L'intrt de mettre  jour cette table est de tenir compte de
                la diminution du nombre des pices au cours de la partie.

  Validation level : 1 - tools.update_pieces
*******************************************************************************/
void update_pieces (void) {

/* ------------------------------- Local data ------------------------------- */

int color;
int x, y;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("update_pieces", true, &update_pieces_calls);
#endif /* DEBUG */

  for (color = BLACK; color <= WHITE; color++) {

    pieces_number[color] = 0;

    for (y = 8; y >= 1; y--) {

      for (x = 1; x <= 8; x++) {

        if (board[x][y].piece) {

          if (board[x][y].color == color) {

            pieces_number[color]++;

            pieces[color][pieces_number[color]].piece    = board[x][y].piece;
            pieces[color][pieces_number[color]].x        = x;
            pieces[color][pieces_number[color]].y        = y;
            pieces[color][pieces_number[color]].captured = false;

            board[x][y].form = pieces_number[color];
          }
        }
        else {

          /* Pour prendre en compte le fait que la position de dpart n'est pas
             toujours la position classique avec toutes les pices : */
          board[x][y].moved = true;
        }
      }
    }
  }

#ifdef DEBUG_HIGH
debug.print (" : ");
for (color = BLACK; color <= WHITE; color++) {
  int i;
  for (i = 1; i <= pieces_number[color]; i++) {
    if (!pieces[color][i].captured) {
      debug.print ("%c%c%d - ",
        PIECE (pieces[color][i].piece),
        pieces[color][i].x + (int) 'a' -1,
        pieces[color][i].y);
    }
  }
}
debug.print ("\n");
#endif /* DEBUG_HIGH */
}
/* End of function : update_pieces */

/*******************************************************************************
  Function : verify_move

  Description : Pour vrifier si un coup au format LAN est lgal dans la
                position courante.

  Validation level : 1 - tools.verify_move
*******************************************************************************/
bool verify_move (char input_lan_move[]) {

/* ------------------------------- Local data ------------------------------- */

char lan_move[6];

int moves_number;

move_s legal_moves[MOVES_LIMIT];

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("verify_move\n", false, &verify_move_calls);
#endif /* DEBUG */

  moves_number = moves.generate (legal_moves, LEGAL_MOVES);

  do {

    tools.convert_move (legal_moves[moves_number], lan_move);

    if (!strcmp (input_lan_move, lan_move)) {

      return true;
    }

  } while (--moves_number);

  return false;
}
/* End of function : verify_move */

/*******************************************************************************
  Object declaration :
*******************************************************************************/

tools_module_s tools = {

  compute_time,
  convert_lan_move,
  convert_move,
  generate_random,
  is_repeated,
  print_board,
  read_input,
  send_output,
  update_pieces,
  verify_move
};
