/*******************************************************************************
  File : uci.c

  Description : Fonctions dfinies par le protocole UCI (Universal Chess
                Interface).
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "data.h"
#include "debug.h"
#include "hash.h"
#include "moves.h"
#include "sissa.h"
#include "tools.h"
#include "uci.h"

/*******************************************************************************
  Macro-instructions :
*******************************************************************************/

/* Pour factoriser l'analyse de la commande go : */
#define parse_command() { \
  pointer = strtok (NULL, " "); \
  if (pointer == NULL) { \
    tools.send_output (true, "ERROR : Missing argument with uci command.\n"); \
    return; \
  } \
}

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int go_calls = 0;
int info_calls = 0;
int isready_calls = 0;
int position_calls = 0;
int quit_calls = 0;
int uci_cmd_calls = 0;
int ucinewgame_calls = 0;
#endif /* DEBUG */

bool uci_infinite;
bool uci_ponder;

int uci_binc;
int uci_btime;
int uci_movetime;
int uci_winc;
int uci_wtime;

int uci_depth;
int uci_mate;
int uci_movestogo;
int uci_nodes;

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : debug_cmd (not used)

  Description : Switch the debug mode of the engine on and off.
*******************************************************************************/
void debug_cmd (void) {
}
/* End of function : debug_cmd */

/*******************************************************************************
  Function : go

  Description : Start calculating on the current position set up with the
                "position" command.

  Validation level : 1 - uci.go
*******************************************************************************/
void go (char string[]) {

/* ------------------------------- Local data ------------------------------- */

const char *pointer;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("go\n", true, &go_calls);
#endif /* DEBUG */

  /* To skip the first string (go) : */
  pointer = strtok (string," ");

  for (pointer = strtok (NULL, " ");
       pointer != NULL;
       pointer = strtok (NULL, " ")) {

    if (false) {

    } else if (!strcmp (pointer, "binc")) {

      parse_command();

      uci_binc = atoi (pointer);

    } else if (!strcmp (pointer, "btime")) {

      parse_command();

      uci_btime = atoi (pointer);

    } else if (!strcmp (pointer, "depth")) {

      parse_command();

      uci_depth = atoi (pointer);

    } else if (!strcmp (pointer, "infinite")) {

      uci_infinite = true;

    } else if (!strcmp (pointer, "mate")) {

      parse_command();

      uci_mate = atoi (pointer);

    } else if (!strcmp (pointer, "movestogo")) {

      parse_command();

      uci_movestogo = atoi (pointer);

    } else if (!strcmp (pointer, "movetime")) {

      parse_command();

      uci_movetime = atoi (pointer);

    } else if (!strcmp (pointer, "nodes")) {

      parse_command();

      uci_nodes = atoi (pointer);

    } else if (!strcmp (pointer, "ponder")) {

      uci_ponder = true;

    } else if (!strcmp (pointer, "searchmoves")) {

      /* Not implemented */

    } else if (!strcmp (pointer, "winc")) {

      parse_command();

      uci_winc = atoi (pointer);

    } else if (!strcmp (pointer, "wtime")) {

      parse_command();

      uci_wtime = atoi (pointer);
    }
  }
}
/* End of function : go */

/*******************************************************************************
  Function : info

  Description : Display the evaluation and others informations from the engine,
                according to the uci conventions in uci mode, otherwise in a
                more human readable format.

  Validation level : 1 - uci.info
*******************************************************************************/
void info (int score) {

/* ------------------------------- Local data ------------------------------- */

double nps;

int elapsed_time;
int i;

char lan_move[STRING_LENGTH];

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("info\n", true, &info_calls);
#endif /* DEBUG */

  elapsed_time = (int) difftime (time (0), start_time);

  if (uci_mode) {

    nps = elapsed_time ? (float) nodes / (float) elapsed_time : 0.0;

    printf ("info depth %d score cp %d time %d nodes %d nps %d pv ",
      depth, score, elapsed_time * 1000, nodes, (int) nps);
  }
  else {

    tools.send_output (true, "   %2d %5.2f %3d %8d %8d : ",
      depth, (float) score / 100, elapsed_time, nodes, qnodes);
  }

  /* Affichage des six premiers coups de la variante principale : */
  for (i = 1; i < pv_length[1] && i <= 6; i++) {

    tools.convert_move (pv[1][i], lan_move);
    tools.send_output (true, "%s ", lan_move);
  }

  tools.send_output (true, "\n");
}
/* End of function : info */

/*******************************************************************************
  Function : isready

  Description : This is used to synchronize the engine with the GUI.

  Validation level : 1 - uci.isready
*******************************************************************************/
void isready (void) {

#ifdef DEBUG
debug.trace ("isready\n", true, &isready_calls);
#endif /* DEBUG */

  /* Indique que le programme est en attente d'ordre de la part de l'interface
     graphique : */
  tools.send_output (true, "readyok\n");
}
/* End of function : isready */

/*******************************************************************************
  Function : ponderhit (not used)

  Description : The user has played the expected move. This will be sent if the
                engine was told to ponder on the same move the user has played.
*******************************************************************************/
void ponderhit (void) {
}
/* End of function : ponderhit */

/*******************************************************************************
  Function : position

  Description : Set up the position described in fenstring on the internal board
                and play the moves on the internal chess board.

  Validation level : 1 - uci.position
*******************************************************************************/
void position (char string[]) {

/* ------------------------------- Local data ------------------------------- */

char lan_move[6];
char *pointer;
char position [STRING_LENGTH];
char *position_moves;

int i;

move_s move;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("position\n", true, &position_calls);
#endif /* DEBUG */

  if (strstr (string, "startpos")) {

    strcpy (position, classic_start_position);
  }
  else if (strstr (string, "fen")) {

    /* ADU : Codage incomplet ... */

    for (i = 13; i < strlen (string); i++) {

      position[i-13] = string[i];
    }
    position[i-13] = '\0';
  }

  /* Initialisation des donnes du programme  partir de la position entre : */
  sissa.init_game (position);

  /* Lecture des coups jous depuis la position ci-dessus : */
  if ((position_moves = strstr (string, "moves")) != NULL) {

    /* To initialize the analyze of string and skip "moves" : */
    pointer = strtok (position_moves, " ");

    for (pointer = strtok (NULL, " ");
         pointer != NULL;
         pointer = strtok (NULL, " ")) {

      strcpy (lan_move, pointer);
      tools.convert_lan_move (lan_move, &move);
      moves.make (move);
    }

    tools.update_pieces ();
  }
}
/* End of function : position */

/*******************************************************************************
  Function : quit

  Description : Quit the program as soon as possible.

  Validation level : 1 - uci.quit
*******************************************************************************/
void quit (int status) {

#ifdef DEBUG
debug.trace ("quit\n", true, &quit_calls);
#endif /* DEBUG */

#ifdef DEBUG
  /* Fermeture des fichiers utiliss pour dboguer le programme : */
  debug.close_debug_files ();

  /* Ecriture d'informations, dans le fichier journal, sur le fonctionnement du
     programme : */
  debug.print_informations ();
  debug.print_statistics ();

  /* Fermeture du fichier journal : */
  fclose (log_file);
#endif /* DEBUG */

  /* Fin du programme : */
  exit (status);
}
/* End of function : quit */

/*******************************************************************************
  Function : registration (not used)

  Description : To try to register an engine.
*******************************************************************************/
void registration (void) {
}
/* End of function : registration */

/*******************************************************************************
  Function : setoption (not used)

  Description : This is sent to the engine when the user wants to change the
                internal parameters of the engine.
*******************************************************************************/
void setoption (void) {
}
/* End of function : setoption */

/*******************************************************************************
  Function : stop (not used)

  Description : Stop calculating the engine as soon as possible.
*******************************************************************************/
void stop (void) {
}
/* End of function : stop */

/*******************************************************************************
  Function : uci_cmd

  Description : Tell the engine to use the universal chess interface.

  Validation level : 1 - uci.uci_cmd
*******************************************************************************/
void uci_cmd (void) {

#ifdef DEBUG
debug.trace ("uci_cmd\n", true, &uci_cmd_calls);
#endif /* DEBUG */

  /* Initialisation de quelques indicateurs en mode UCI : */
  forced_mode = true;
  show_board  = false;
  uci_mode    = true;

  /* Prsentation du programme : */
  tools.send_output (true, "id name %s %s\n", PROGRAM_NAME, PROGRAM_VERSION);
  tools.send_output (true, "id author %s\n", PROGRAM_AUTHOR);
  tools.send_output (true, "uciok\n");
}
/* End of function : uci_cmd */

/*******************************************************************************
  Function : ucinewgame

  Description : This is sent to the engine when the next search (started with
                "position" and "go") will be from a different game.

  Validation level : 1 - uci.ucinewgame
*******************************************************************************/
void ucinewgame (void) {

#ifdef DEBUG
debug.trace ("ucinewgame\n", true, &ucinewgame_calls);
#endif /* DEBUG */

  /* Initialisation d'une nouvelle partie  partir de la position classique : */
  sissa.init_game (classic_start_position);

  /* Temps de base pour chaque coup : */
  base_time = 0;
}
/* End of function : ucinewgame */

/*******************************************************************************
  Object declaration :
*******************************************************************************/

uci_module_s uci = {

  debug_cmd,
  go,
  info,
  isready,
  ponderhit,
  position,
  quit,
  registration,
  setoption,
  stop,
  uci_cmd,
  ucinewgame
};
